#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Stephan Krause <stephan.krause@eox.at>
#          Stephan Meissl <stephan.meissl@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2012 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

"""
WSGI config for EOxServer's {{ project_name }} instance.

This module contains the WSGI application used by Django's development server
and any production WSGI deployments. It should expose a module-level variable
named ``application``. Django's ``runserver`` and ``runfcgi`` commands discover
this application via the ``WSGI_APPLICATION`` setting.

Usually you will have the standard Django WSGI application here, but it also
might make sense to replace the whole Django WSGI application with a custom one
that later delegates to the Django one. For example, you could introduce WSGI
middleware here, or combine a Django application with an application of another
framework.

"""
import os
import sys

path = "{{ project_directory }}"
if path not in sys.path:
    sys.path.append(path)

# NOTE: The Apache mod_wsgi, by default, shares the enviroment variables 
#       between different WSGI apps which leads to conflicts between
#       multiple EOxServer instance. Therefore we cannot rely on the 
#       DJANGO_SETTINGS_MODULE enviromental variable we must always set it
#       to the proper value.
#os.environ.setdefault("DJANGO_SETTINGS_MODULE", "{{ project_name }}.settings")
os.environ["DJANGO_SETTINGS_MODULE"] = "{{ project_name }}.settings"

# Initialize the EOxServer component system.
import eoxserver.core
eoxserver.core.initialize()

# This application object is used by any WSGI server configured to use this
# file. This includes Django's development server, if the WSGI_APPLICATION
# setting points here.
from django.core.wsgi import get_wsgi_application
application = get_wsgi_application()

# Apply WSGI middleware here.
# from helloworld.wsgi import HelloWorldApplication
# application = HelloWorldApplication(application)



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from eoxserver.core.decoders import kvp, xml
from eoxserver.services.ows.wps.util import get_processes
from eoxserver.services.ows.wps.v10.encoders import WPS10CapabilitiesXMLEncoder
from eoxserver.services.ows.wps.v10.util import nsmap


class WPS10GetCapabilitiesHandler(object):
    """ WPS 1.0 GetCapabilities service handler. """

    service = "WPS"
    versions = ("1.0.0",)
    request = "GetCapabilities"
    methods = ['GET', 'POST']

    def handle(self, request):
        """ Handle HTTP request. """
        encoder = WPS10CapabilitiesXMLEncoder()
        return encoder.serialize(encoder.encode_capabilities(get_processes()))


class WPS10GetCapabilitiesKVPDecoder(kvp.Decoder):
    """ WPS 1.0 GetCapabilities HTTP/GET KVP request decoder. """
    #pylint: disable=too-few-public-methods
    #acceptversions = kvp.Parameter(type=typelist(str, ","), num="?")
    language = kvp.Parameter(num="?")


class WPS10GetCapabilitiesXMLDecoder(xml.Decoder):
    """ WPS 1.0 DescribeProcess HTTP/POST XML request decoder. """
    #pylint: disable=too-few-public-methods
    #acceptversions = xml.Parameter("/ows:AcceptVersions/ows:Version/text()", num="*")
    language = xml.Parameter("/ows:AcceptLanguages/ows:Language/text()", num="*")
    namespaces = nsmap



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


import logging
from itertools import chain

from django.db.models import Q
from django.utils.six import MAXSIZE

from eoxserver.core.config import get_eoxserver_config
from eoxserver.core.decoders import xml, kvp, typelist, enum
from eoxserver.render.coverage import objects
from eoxserver.resources.coverages import models
from eoxserver.services.ows.wcs.v20.util import (
    nsmap, SectionsMixIn, parse_subset_kvp, parse_subset_xml
)
from eoxserver.services.ows.wcs.v20.encoders import WCS20EOXMLEncoder
from eoxserver.services.ows.common.config import WCSEOConfigReader
from eoxserver.services.subset import Subsets, Trim
from eoxserver.services.exceptions import (
    NoSuchDatasetSeriesOrCoverageException, InvalidSubsettingException
)


logger = logging.getLogger(__name__)


class WCS20DescribeEOCoverageSetHandler(object):
    service = "WCS"
    versions = ("2.0.0", "2.0.1")
    methods = ['GET', 'POST']
    request = "DescribeEOCoverageSet"

    index = 20

    def get_decoder(self, request):
        if request.method == "GET":
            return WCS20DescribeEOCoverageSetKVPDecoder(request.GET)
        elif request.method == "POST":
            return WCS20DescribeEOCoverageSetXMLDecoder(request.body)

    @property
    def constraints(self):
        reader = WCSEOConfigReader(get_eoxserver_config())
        return {
            "CountDefault": reader.paging_count_default
        }

    def handle(self, request):
        decoder = self.get_decoder(request)
        eo_ids = decoder.eo_ids

        containment = decoder.containment
        if not containment:
            containment = "overlaps"

        count_default = self.constraints["CountDefault"]
        count = decoder.count
        if count_default is not None:
            count = min(count, count_default)

        try:
            subsets = Subsets(
                decoder.subsets,
                crs="http://www.opengis.net/def/crs/EPSG/0/4326",
                allowed_types=Trim
            )
        except ValueError as e:
            raise InvalidSubsettingException(str(e))

        # check whether the DatasetSeries and CoverageDescriptions sections are
        # included
        inc_dss_section = decoder.section_included("DatasetSeriesDescriptions")
        inc_cov_section = decoder.section_included("CoverageDescriptions")

        if len(eo_ids) == 0:
            raise

        # fetch the objects directly referenced by EOID
        eo_objects = models.EOObject.objects.filter(
            identifier__in=eo_ids
        ).select_subclasses()

        # check if all EOIDs are available
        available_ids = set(eo_object.identifier for eo_object in eo_objects)
        failed = [
            eo_id for eo_id in eo_ids if eo_id not in available_ids
        ]

        # fail when some objects are not available
        if failed:
            raise NoSuchDatasetSeriesOrCoverageException(failed)

        # split list of objects into Collections, Products and Coverages
        collections = []
        mosaics = []
        products = []
        coverages = []

        for eo_object in eo_objects:
            if isinstance(eo_object, models.Collection):
                collections.append(eo_object)
            elif isinstance(eo_object, models.Mosaic):
                mosaics.append(eo_object)
            elif isinstance(eo_object, models.Product):
                products.append(eo_object)
            elif isinstance(eo_object, models.Coverage):
                coverages.append(eo_object)

        filters = subsets.get_filters(containment=containment)

        # get a QuerySet of all dataset series, directly or indirectly referenced
        all_dataset_series_qs = models.EOObject.objects.filter(
            Q(  # directly referenced Collections
                collection__isnull=False,
                identifier__in=[
                    collection.identifier for collection in collections
                ],
            ) |
            Q(  # directly referenced Products
                product__isnull=False,
                identifier__in=[product.identifier for product in products],
            ) |
            Q(  # Products within Collections
                product__isnull=False,
                product__collections__in=collections,
                **filters
            )
        )

        if inc_dss_section:
            dataset_series_qs = all_dataset_series_qs[:count]
        else:
            dataset_series_qs = models.EOObject.objects.none()

        # Allow metadata queries on coverage itself or on the
        # parent product if available
        parent_product_filters = []
        for key, value in filters.items():
            prop = key.partition('__')[0]
            parent_product_filters.append(
                Q(**{
                    key: value
                }) | Q(**{
                    '%s__isnull' % prop: True,
                    'coverage__parent_product__%s' % key: value
                })
            )

        # get a QuerySet for all Coverages, directly or indirectly referenced
        all_coverages_qs = models.EOObject.objects.filter(
            *parent_product_filters
        ).filter(
            Q(  # directly referenced Coverages
                identifier__in=[
                    coverage.identifier for coverage in coverages
                ]
            ) |
            Q(  # Coverages within directly referenced Products
                coverage__parent_product__in=products,
            ) |
            Q(  # Coverages within indirectly referenced Products
                coverage__parent_product__collections__in=collections
            ) |
            Q(  # Coverages within directly referenced Collections
                coverage__collections__in=collections
            ) |
            Q(  # Coverages within directly referenced Collections
                coverage__mosaics__in=mosaics
            ) |
            Q(  # directly referenced Mosaics
                identifier__in=[
                    mosaic.identifier for mosaic in mosaics
                ]
            ) |
            Q(  # Mosaics within directly referenced Collections
                mosaic__collections__in=collections
            )
        ).select_subclasses(models.Coverage, models.Mosaic)

        all_coverages_qs = all_coverages_qs.order_by('identifier')

        # check if the CoverageDescriptions section is included. If not, use an
        # empty queryset
        if inc_cov_section:
            coverages_qs = all_coverages_qs
        else:
            coverages_qs = models.Coverage.objects.none()

        # limit coverages according to the number of dataset series
        coverages_qs = coverages_qs[:max(
            0, count - dataset_series_qs.count() - len(mosaics)
        )]

        # compute the number of all items that would match
        number_matched = all_coverages_qs.count() + all_dataset_series_qs.count()

        # create an encoder and encode the result
        encoder = WCS20EOXMLEncoder()
        return (
            encoder.serialize(
                encoder.encode_eo_coverage_set_description(
                    dataset_series_set=[
                        objects.DatasetSeries.from_model(eo_object)
                        for eo_object in dataset_series_qs
                    ],
                    coverages=[
                        objects.from_model(coverage)
                        for coverage in coverages_qs
                    ],
                    number_matched=number_matched
                ), pretty_print=True
            ),
            encoder.content_type
        )


def pos_int(value):
    value = int(value)
    if value < 0:
        raise ValueError("Negative values are not allowed.")
    return value


containment_enum = enum(
    ("overlaps", "contains"), False
)

sections_enum = enum(
    ("DatasetSeriesDescriptions", "CoverageDescriptions", "All"), False
)

class WCS20DescribeEOCoverageSetKVPDecoder(kvp.Decoder, SectionsMixIn):
    eo_ids      = kvp.Parameter("eoid", type=typelist(str, ","), num=1, locator="eoid")
    subsets     = kvp.Parameter("subset", type=parse_subset_kvp, num="*")
    containment = kvp.Parameter(type=containment_enum, num="?")
    count       = kvp.Parameter(type=pos_int, num="?", default=MAXSIZE)
    sections    = kvp.Parameter(type=typelist(sections_enum, ","), num="?")


class WCS20DescribeEOCoverageSetXMLDecoder(xml.Decoder, SectionsMixIn):
    eo_ids      = xml.Parameter("wcseo:eoId/text()", num="+", locator="eoid")
    subsets     = xml.Parameter("wcs:DimensionTrim", type=parse_subset_xml, num="*")
    containment = xml.Parameter("wcseo:containment/text()", type=containment_enum, locator="containment")
    count       = xml.Parameter("@count", type=pos_int, num="?", default=MAXSIZE, locator="count")
    sections    = xml.Parameter("wcseo:sections/wcseo:section/text()", type=sections_enum, num="*", locator="sections")

    namespaces = nsmap



#-------------------------------------------------------------------------------
#
#  Execute - XML decoder
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#          Martin Paces <martin.paces@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from lxml import etree
from eoxserver.core.decoders import xml
from eoxserver.services.ows.wps.exceptions import InvalidParameterValue
from eoxserver.services.ows.wps.parameters import (
    InputData, InputReference, Output, ResponseDocument, RawDataOutput
)
from eoxserver.services.ows.wps.v10.util import nsmap, ns_wps, ns_ows, ns_xlink
from eoxserver.services.ows.wps.v10.execute_decoder_common import (
    parse_bool,
)


def _parse_input(element):
    """ Parse one data input element. """
    id_ = element.findtext("./"+ns_ows("Identifier"))
    title = element.findtext("./"+ns_ows("Title"))
    abstract = element.findtext("./"+ns_ows("Abstract"))

    if id_ is None:
        raise ValueError("Missing the mandatory input identifier!")

    elem_ref = element.find("./"+ns_wps("Reference"))
    elem_data = element.find("./"+ns_wps("Data"))

    if elem_ref is not None:
        value = _parse_input_reference(elem_ref, id_, title, abstract)

    elif elem_data is not None:
        if len(elem_data) != 1:
            raise ValueError("Invalid input content of the 'wps:Data' element!")
        # pylint: disable=redefined-variable-type
        value = _parse_input_data(elem_data[0], id_, title, abstract)

    return id_, value


def _parse_response_form(elem_rform):
    """ Parse ResponseForm element holding either ResponseDocument or
    RawDataOutput elements.
    """
    elem_rdoc = elem_rform.find("./"+ns_wps("ResponseDocument"))
    if elem_rdoc is not None:
        rdoc = ResponseDocument(
            lineage=parse_bool(elem_rdoc.attrib.get("lineage")),
            status=parse_bool(elem_rdoc.attrib.get("status")),
            store_response=parse_bool(
                elem_rdoc.attrib.get("storeExecuteResponse")
            ),
        )
        for elem in elem_rdoc.iterfind("./"+ns_wps("Output")):
            id_ = elem.findtext(ns_ows("Identifier"))
            title = elem.findtext(ns_ows("Title"))
            abstr = elem.findtext(ns_ows("Abstract"))
            rdoc.set_output(_create_output(id_, elem.attrib, title, abstr))
        return rdoc

    elem_rawout = elem_rform.find("./"+ns_wps("RawDataOutput"))
    if elem_rawout is not None:
        id_ = elem_rawout.findtext(ns_ows("Identifier"))
        return RawDataOutput(_create_output(id_, elem_rawout.attrib))

    raise InvalidParameterValue('Invalid ResponseForm!', 'ResponseForm')


def _parse_input_reference(elem, identifier, title, abstract):
    """ Parse one input item passed as a reference. """
    href = elem.attrib.get(ns_xlink("href"))
    if href is None:
        raise ValueError("Missing the mandatory 'xlink:href' attribute!")

    body = elem.findtext("./"+ns_wps("Body"))
    elem_tmp = elem.find("./"+ns_wps("BodyReference"))
    body_href = elem_tmp.attrib.get(ns_xlink("href")) if elem_tmp else None

    headers = dict(
        (header.attrib["key"], header.attrib["value"])
        for header in elem.iterfind("./"+ns_wps("Header"))
    )

    return InputReference(
        identifier, title, abstract,
        href, headers, body,
        elem.attrib.get("method", "GET"),
        elem.attrib.get("mimeType"),
        elem.attrib.get("encoding"),
        elem.attrib.get("schema"),
        body_href,
    )


def _parse_input_data(elem, identifier, title, abstract):
    """ Parse one input item value. """
    if elem.tag == ns_wps("LiteralData"):
        args = _parse_input_literal(elem)
    elif elem.tag == ns_wps("BoundingBoxData"):
        args = _parse_input_bbox(elem)
    elif elem.tag == ns_wps("ComplexData"):
        args = _parse_input_complex(elem)
    else:
        raise ValueError("Invalid input content of the 'wps:Data' element!")
    return InputData(
        identifier=identifier, title=title, abstract=abstract, **args
    )


def _parse_input_literal(elem):
    """ Parse one literal data input item value. """
    args = {}
    args['data'] = elem.text or ""
    args['uom'] = elem.attrib.get("uom")
    return args


def _parse_input_bbox(elem):
    """ Parse one bounding-box input item value. """
    args = {}
    lower_corner = elem.findtext("./"+ns_ows("LowerCorner"))
    upper_corner = elem.findtext("./"+ns_ows("UpperCorner"))
    if lower_corner is None or upper_corner is None:
        raise ValueError("Invalid 'wps:BoundingBoxData' element!")
    args['data'] = (lower_corner, upper_corner, elem.attrib.get("crs"))
    return args


def _parse_input_complex(elem):
    """ Parse one complex data input item value. """
    args = {}
    if len(elem):
        args['data'] = etree.tostring(
            elem[0], pretty_print=False, xml_declaration=True, encoding="utf-8"
        )
    else:
        args['data'] = elem.text or ""
    args['mime_type'] = elem.attrib.get("mimeType")
    args['encoding'] = elem.attrib.get("encoding")
    args['schema'] = elem.attrib.get("schema")
    return args


def _create_output(identifier, attrs, title=None, abstract=None):
    """ Create one Output object.from the parsed id and attributes. """
    return Output(
        identifier, title, abstract, attrs.get("uom"),
        attrs.get("crs"), attrs.get("mimeType"), attrs.get("encoding"),
        attrs.get("schema"), parse_bool(attrs.get("asReference"))
    )


class WPS10ExecuteXMLDecoder(xml.Decoder):
    """ WPS 1.0 POST/XML Execute request decoder class. """
    identifier = xml.Parameter("ows:Identifier/text()")
    _inputs = xml.Parameter(
        "wps:DataInputs/wps:Input", type=_parse_input, num="*", default=[]
    )
    _response_form = xml.Parameter(
        "wps:ResponseForm", type=_parse_response_form, num="?"
    )

    @property
    def response_form(self):
        """ Get the unified response form object. """
        resp_form = self._response_form
        return resp_form if resp_form is not None else ResponseDocument()

    @property
    def inputs(self):
        """ Get the raw data inputs as a dictionary. """
        return dict(self._inputs)

    namespaces = nsmap



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Martin Paces <martin.paces@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

"""
This module contains implementation of MIME multipart packing and unpacking
utilities.

The main benefit of the utilities over other methods of mutipart handling
is that the functions of this module do not manipulate the input data
buffers and especially avoid any unnecessary data copying.
"""
from django.utils.six import b

def capitalize(header_name):
    """ Capitalize header field name. Eg., 'content-type' is capilalized to
    'Content-Type'.

    .. deprecated:: 0.4
    """
    return "-".join([f.capitalize() for f in header_name.split("-")])

# local alias to prevent conflict with local variable
__capitalize = capitalize


def getMimeType(content_type):
    """ Extract MIME-type from Content-Type string and convert it to
    lower-case.

    .. deprecated:: 0.4
    """
    return content_type.partition(";")[0].strip().lower()


def getMultipartBoundary(content_type):
    """ Extract boundary string from mutipart Content-Type string.

    .. deprecated:: 0.4
    """

    for opt in content_type.split(";")[1:]:
        key, _, val = opt.partition("=")
        if key.strip().lower() == "boundary":
            return val.strip()

    raise ValueError(
        "failed to extract the mutipart boundary string! content-type: %s"
        % content_type
    )


def mpPack(parts, boundary):
    """
Low-level memory-friendly MIME multipart packing.

Note: The data payload is passed untouched and no transport encoding
of the payload is performed.

Inputs:

 - parts - list of part-tuples, each tuple shall have two elements
    the header list and (string) payload. The header itsels should be
    a sequence of key-value pairs (tuples).

 - boundary - boundary string

Ouput:

 - list of strings (which can be directly passsed as a Django response content)

.. deprecated:: 0.4
    """

    # empty multipart package
    pack = ["--%s" % boundary]

    for header, data in parts:

        # pack header
        for key, value in header:
            pack.append("\r\n%s: %s" % (key, value))

        # terminate header
        pack.append("\r\n\r\n")

        # append data
        pack.append(data)

        # terminate partition
        pack.append("\r\n--%s" % boundary)

    #terminate package
    pack.append("--")

    # return package
    return pack


def mpUnpack(cbuffer, boundary, capitalize=False):
    """
Low-level memory-friendly MIME multipart unpacking.

Note: The payload of the multipart package data is neither modified nor copied.
No decoding of the transport encoded payload is performed.

Note: The subroutine does not unpack any nested mutipart content.

Inputs:

 - ``cbuffer`` - character buffer (string) containing the
   the header list and (string) payload. The header itsels should be
   a sequence of key-value pairs (tuples).

 - ``boundary`` - boundary string

 - ``capitalize`` - by default the header keys are converted to lower-case
   (e.g., 'content-type').
   To capitalize the names (e.g., 'Content-Type') set this option to true.

Output:

 - list of parts - each part is a tuple of the header dictionary,
   payload ``cbuffer`` offset and payload size.

.. deprecated:: 0.4
    """

    def findBorder(offset=0):

        delim = "--%s" % boundary if offset == 0 else "\n--%s" % boundary

        # boundary offset (end of last data)
        idx0 = cbuffer.find(delim, offset)

        if idx0 < 0:
            raise ValueError("Boundary cannot be found!")

        # header offset
        idx1 = idx0 + len(delim)

        # nescessary check to be able to safely check two following characters
        if len(cbuffer[idx1:]) < 2:
            raise ValueError("Buffer too short!")

        # check the leading CR character
        if idx0 > 0 and cbuffer[idx0-1] == "\r":
            idx0 -= 1

        # check the terminating sequence
        if cbuffer[idx1:(idx1+2)] == "--":
            return idx0, idx1+2, -1

        # look-up double endl-line (data offset)
        tmp = idx1

        while True:

            tmp = 1 + cbuffer.find("\n", tmp)

            if tmp < 1:
                raise ValueError(
                    "Cannot find payload's a double new-line separator!"
                )

            # is it followed by new line?

            elif cbuffer[tmp:(tmp+2)] == "\r\n":
                idx2 = tmp + 2
                break

            elif cbuffer[tmp:(tmp+1)] == "\n":
                idx2 = tmp + 1
                break

            # otherwise continue to lookup
            continue

        # adjust the data offset (separator must be followed by new-line)
        if cbuffer[idx1:(idx1+2)] == "\r\n":
            idx1 += 2
        elif cbuffer[idx1:(idx1+1)] == "\n":
            idx1 += 1
        else:
            raise ValueError("Boundary is not followed by a new-line!")

        return idx0, idx1, idx2

    #--------------------------------------------------------------------------
    # auxiliary nested functions formating header names

    # capitalize header name
    def unpackCC(v):
        key, _, val = v.partition(b(":"":"))
        return __capitalize(key.strip()), val.strip()

    # header name all lower
    def unpackLC(v):
        key, _, val = v.partition(b(":"))
        return key.strip().lower(), val.strip()

    # filter function rejecting entries with blank keys
    def noblank(tup):
        (k, v) = tup
        return bool(k)

    #--------------------------------------------------------------------------

    # get the offsets
    # off = (<last payload end>,<header start>,<payload start>)
    # negative <payload start> means terminating boundary

    try:

        off = findBorder()
        offsets = [off]

        while off[1] < off[2]:
            off = findBorder(off[2])
            offsets.append(off)

    except ValueError as e:
        raise Exception(
            "The buffer is not a valid MIME multi-part message! Reason: %s"
            % e.message
        )

    # process the parts
    parts = []
    for of0, of1 in zip(offsets[:-1], offsets[1:]):

        # get the http header with <LF> line ending
        tmp = cbuffer[of0[1]:of0[2]].replace("\r\n", "\n")[:-2].split("\n")

        # unpack header
        header = dict(
            filter(noblank,
                map((unpackLC, unpackCC)[capitalize], tmp)
            )
        )

        # get the header and payload offset and size
        parts.append((header, of0[2], of1[0]-of0[2]))

    return parts


CRLF = b"\r\n"
CRLFCRLF = b"\r\n\r\n"


def get_substring(data, boundary, offset, end):
    """ Retrieves the substring of ``data`` until the next ``boundary`` from a
    given offset to a until ``end``.
    """
    index = data.find(boundary, offset, end)
    if index == -1:
        return None, None
    return data[offset:index], index + len(boundary)


def parse_parametrized_option(string):
    """ Parses a parametrized options string like
    'base;option=value;otheroption=othervalue'.

    :returns: the base string and a :class:`dict` with all parameters
    """
    parts = string.split(b";")
    params = dict(
        param.strip().split(b"=", 1) for param in parts[1:]
    )
    return parts[0], params


def capitalize_header(key):
    """ Returns a capitalized version of the header line such as
    'content-type' -> 'Content-Type'.
    """

    return b"-".join([
        item
        if item.decode()[0].isupper() else
        (item.decode()[0].upper() + item.decode()[1:]).encode('ascii')
        for item in key.split(b"-")
    ])


def iterate(data, offset=0, end=None, headers=None):
    """ Efficient generator function to iterate over a single- or multipart
        message. I yields tuples in the shape (``headers``, ``data``), where
        headers is a ``dict`` and data a buffer object, referencing the subset
        of the original content. In case of multipart messages, the multipart
        headers are yielded beforehand, with an empty string as data.

        The `offset` parameter specifies the offset index to the start of the
        data. This is mostly used in the recursive call. The same applies to the
        `end` parameter.

        The `headers` parameter specifies that the header section of the
        response was already read, and the headers are now entailed in the given
        dict. If this parameter is omitted, the headers are read from the
        stream.
    """

    # check if the headers need to be parsed.
    if not headers:
        # read the header bytes from the string and get the new offset.
        header_bytes, offset = get_substring(data, CRLFCRLF, offset, end)

        # check if no data could be extracted
        if (header_bytes, offset) == (None, None):
            return

        # parse the headers into a dict
        headers = {}
        for line in header_bytes.split(CRLF):
            key, _, value = line.partition(b":")
            headers[capitalize_header(key.strip())] = value.strip()

    # get the content type
    content_type, params = parse_parametrized_option(
        headers.get(b"Content-Type", b"")
    )

    # check if this is a multipart
    if content_type.startswith(b"multipart"):
        # if this is a multipart, yield only its headers and an empty string
        yield headers, memoryview(b"")

        # parse the boundary and find the final index of all multiparts
        boundary = b"%s--%s" % (CRLF, params[b"boundary"])
        end_boundary = b"%s--" % boundary

        sub_end = data.find(end_boundary)
        if sub_end == -1:
            raise ValueError("Could not find multipart end.")

        # get the first part of this multipart
        sub_offset = data.find(boundary, offset, sub_end)
        if sub_offset == -1:
            raise ValueError("Could not find boundary.")

        # iterate over all parts until we reach the end of the multipart
        while sub_offset < sub_end:
            sub_offset += len(boundary) + 1
            sub_stop = data.find(boundary, sub_offset, sub_end)

            sub_stop = sub_stop if sub_stop > -1 else sub_end

            # recursive function call
            for sub_headers, sub_data in iterate(data, sub_offset, sub_stop):
                yield sub_headers, sub_data

            sub_offset = sub_stop

    else:
        # in case we have a single part, just yield the headers and a buffer
        # pointing to a substring of the original data stream.
        if end is not None:
            yield headers, memoryview(data)[offset:end]
        else:
            yield headers, memoryview(data)[offset:]



#-------------------------------------------------------------------------------
#
#  Execute - KVP decoder
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#          Martin Paces <martin.paces@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from django.utils.six.moves.urllib.parse import unquote_plus

from eoxserver.core.decoders import kvp
from eoxserver.services.ows.wps.parameters import (
    InputData, InputReference, Output, ResponseDocument, RawDataOutput
)
from eoxserver.services.ows.wps.v10.execute_decoder_common import (
    parse_bool,
)


def _parse_inputs(raw_string):
    """ Parse DataInputs value. """
    inputs = {}
    for item in raw_string.split(";"):
        id_, value, param = _parse_param(item)
        href = param.get("href") or param.get("xlink:href")
        if href is not None:
            input_ = InputReference(
                href=href,
                identifier=id_,
                mime_type=param.get("mimeType"),
                encoding=param.get("encoding"),
                schema=param.get("schema")
            )
        else:
            #NOTE: KVP Bounding box cannot be safely detected and parsed.
            input_ = InputData( # pylint: disable=redefined-variable-type
                data=value,
                identifier=id_,
                uom=param.get("uom"),
                mime_type=param.get("mimeType"),
                encoding=param.get("encoding"),
                schema=param.get("schema"),
                asurl=True,
            )
        inputs[id_] = input_
    return inputs


def _parse_param(raw_string):
    """ Parse one input or output item. """
    items = (item.partition('=') for item in raw_string.split("@"))
    attr = {}
    id_, dlm, data = next(items)
    id_ = unquote_plus(id_)
    data = unquote_plus(data) if dlm else None
    for key, dlm, value in items:
        attr[unquote_plus(key)] = unquote_plus(value) if dlm else None
    return id_, data, attr


def _parse_outputs(raw_string):
    """ Parse ResponseDocument parameter. """
    outputs = []
    for output in raw_string.split(";"):
        outputs.append(_create_output(*_parse_param(output)))
    return outputs


def _parse_raw_output(raw_string):
    """ Parse RawDataOutput parameter. """
    return RawDataOutput(_create_output(*_parse_param(raw_string)))


def _create_output(identifier, _, attrs):
    """ Create one Output object from the parsed identifier and attributes. """
    return Output(
        identifier, None, None, attrs.get("uom"),
        attrs.get("crs"), attrs.get("mimeType"), attrs.get("encoding"),
        attrs.get("schema"), parse_bool(attrs.get("asReference"))
    )


def parse_query_string(query_string):
    """ Parse URL query string preserving the URL-encoded
    DataInputs, ResponseDocument, and RawDataOutput WPS Execute parameters.
    Note that the standard parser URL-decodes the parameter values and, in cases
    when, e.g., a data input contains an percent-encoded separator
    ('%40' vs. '@') the encoded and non-encoded delimiters cannot
    be distinguished ('@' vs. '@') and the correct parsing cannot be guaranteed.
    """
    unescaped = set(('datainputs', 'responsedocument', 'rawdataoutput'))
    return dict(
        (key, value if key.lower() in unescaped else unquote_plus(value))
        for key, value in (
            (unquote_plus(key), value) for key, _, value in (
                item.partition('=') for item in query_string.split('&')
            )
        )
    )


class WPS10ExecuteKVPDecoder(kvp.Decoder):
    """ WPS 1.0 Execute HTTP/GET KVP request decoder. """
    #pylint: disable=too-few-public-methods
    identifier = kvp.Parameter()
    inputs = kvp.Parameter(
        "DataInputs", type=_parse_inputs, num="?", default={}
    )
    outputs = kvp.Parameter(
        "ResponseDocument", type=_parse_outputs, num="?", default=[]
    )
    raw_response = kvp.Parameter(
        "RawDataOutput", type=_parse_raw_output, num="?"
    )
    status = kvp.Parameter(
        "status", type=parse_bool, num="?", default=False
    )
    lineage = kvp.Parameter(
        "lineage", type=parse_bool, num="?", default=False
    )
    store_response = kvp.Parameter(
        "storeExecuteResponse", type=parse_bool, num="?", default=False
    )

    @property
    def response_form(self):
        """ Get response unified form parsed either from ResponseDocument or
        RawDataOutput parameters.
        """
        raw_response = self.raw_response
        if raw_response:
            return raw_response

        resp_doc = ResponseDocument(
            lineage=self.lineage,
            status=self.status,
            store_response=self.store_response
        )
        for output in self.outputs: # pylint: disable=not-an-iterable
            resp_doc.set_output(output)
        return resp_doc



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Stephan Krause <stephan.krause@eox.at>
#          Stephan Meissl <stephan.meissl@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

"""
Django settings for EOxServer's {{ project_name }} instance.

For more information on this file, see
https://docs.djangoproject.com/en/1.11/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/1.11/ref/settings/
"""

import os
from os.path import join, abspath, dirname

PROJECT_DIR = dirname(abspath(__file__))
PROJECT_URL_PREFIX = ''

#TEST_RUNNER = 'eoxserver.testing.core.EOxServerTestRunner'

DEBUG = True
TEMPLATE_DEBUG = DEBUG

ADMINS = (
    # ('EOX', 'office@eox.at'),
)

MANAGERS = ADMINS

# Configure which database to use.
db_type = os.environ.get('DB')
if db_type == 'postgis':
    DATABASES = {
        'default': {
            'ENGINE': 'django.contrib.gis.db.backends.postgis',
            'HOST': os.environ['DB_HOST'],
            'NAME': os.environ['DB_NAME'],
            'USER': os.environ['DB_USER'],
            'PASSWORD': os.environ['DB_PW'],
        }
    }
elif db_type == 'spatialite':
    spatialite_path = os.environ.get('SPATIALITE_PATH', 'data/config.sqlite')
    DATABASES = {
        'default': {
            'ENGINE': 'django.contrib.gis.db.backends.spatialite',
            'NAME': join(PROJECT_DIR, spatialite_path),
        }
    }

    SPATIALITE_SQL = join(PROJECT_DIR, 'data/init_spatialite-2.3.sql')
    SPATIALITE_LIBRARY_PATH = 'mod_spatialite.so'

# Hosts/domain names that are valid for this site; required if DEBUG is False
# See https://docs.djangoproject.com/en/1.5/ref/settings/#allowed-hosts
ALLOWED_HOSTS = []

# Local time zone for this installation. Choices can be found here:
# http://en.wikipedia.org/wiki/List_of_tz_zones_by_name
# although not all choices may be available on all operating systems.
# In a Windows environment this must be set to your system time zone.
# Note that this is the time zone to which Django will convert all
# dates/times -- not necessarily the timezone of the server.
# If you are using UTC (Zulu) time zone for your data (e.g. most
# satellite imagery) it is highly recommended to use 'UTC' here. Otherwise
# you will encounter time-shifts between your data, search request & the
# returned results.
TIME_ZONE = 'UTC'

# Language code for this installation. All choices can be found here:
# http://www.i18nguy.com/unicode/language-identifiers.html
LANGUAGE_CODE = 'en-us'

SITE_ID = 1

# If you set this to False, Django will make some optimizations so as not
# to load the internationalization machinery.
USE_I18N = True

# If you set this to False, Django will not format dates, numbers and
# calendars according to the current locale.
USE_L10N = True

# If you set this to False, Django will not use timezone-aware datetimes.
USE_TZ = True

# Absolute filesystem path to the directory that will hold user-uploaded files.
# Example: "/var/www/example.com/media/"
MEDIA_ROOT = ''

# URL that handles the media served from MEDIA_ROOT. Make sure to use a
# trailing slash.
# Examples: "http://example.com/media/", "http://media.example.com/"
MEDIA_URL = ''

# Absolute path to the directory static files should be collected to.
# Don't put anything in this directory yourself; store your static files
# in apps' "static/" subdirectories and in STATICFILES_DIRS.
# Example: "/var/www/example.com/static/"
STATIC_ROOT = join(PROJECT_DIR, 'static')

# URL prefix for static files.
# Example: "http://example.com/static/", "http://static.example.com/"
STATIC_URL = '/{{ project_name }}_static/'

# Additional locations of static files
STATICFILES_DIRS = (
    # Put strings here, like "/home/html/static" or "C:/www/django/static".
    # Always use forward slashes, even on Windows.
    # Don't forget to use absolute paths, not relative paths.
)

# List of finder classes that know how to find static files in
# various locations.
STATICFILES_FINDERS = (
    'django.contrib.staticfiles.finders.FileSystemFinder',
    'django.contrib.staticfiles.finders.AppDirectoriesFinder',
#    'django.contrib.staticfiles.finders.DefaultStorageFinder',
)

# Make this unique, and don't share it with anybody.
SECRET_KEY = '{{ secret_key }}'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    # 'django.contrib.auth.middleware.SessionAuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    # # For management of the per/request cache system.
    # 'eoxserver.backends.middleware.BackendsCacheMiddleware',
]

MIDDLEWARE_CLASSES = (
    'django.middleware.common.CommonMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
# Commented because of POST requests:    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    # Uncomment the next line for simple clickjacking protection:
    # 'django.middleware.clickjacking.XFrameOptionsMiddleware',

    # For management of the per/request cache system.
    'eoxserver.backends.middleware.BackendsCacheMiddleware',
)

ROOT_URLCONF = '{{ project_name }}.urls'

# Python dotted path to the WSGI application used by Django's runserver.
WSGI_APPLICATION = '{{ project_name }}.wsgi.application'

INSTALLED_APPS = (
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.sites',
    'django.contrib.messages',
    'django.contrib.gis',
    'django.contrib.staticfiles',
    # Enable the admin:
    'django.contrib.admin',
    # Enable admin documentation:
    'django.contrib.admindocs',
    # Enable the databrowse:
    #'django.contrib.databrowse',
    # Enable for better schema and data-migrations
    #'south',
    # Enable for debugging
    #'django_extensions',
    # Enable EOxServer:
    'eoxserver.core',
    'eoxserver.services',
    'eoxserver.resources.coverages',
    'eoxserver.backends',
    'eoxserver.testing',
    'eoxserver.webclient'
)


# The configured EOxServer components. Components add specific functionality
# to the EOxServer and must adhere to a given interface. In order to activate
# a component, its module must be included in the following list or imported at
# some other place. To help configuring all required components, each module
# path can end with either a '*' or '**'. The single '*' means that all direct
# modules in the package will be included. With the double '**' a recursive
# search will be done.
COMPONENTS = (
    # not used anymore
)


# A sample logging configuration. The only tangible logging
# performed by this configuration is to send an email to
# the site admins on every HTTP 500 error when DEBUG=False.
# See http://docs.djangoproject.com/en/dev/topics/logging for
# more details on how to customize your logging configuration.
LOGGING = {
    'version': 1,
    'disable_existing_loggers': True,
    'filters': {
        'require_debug_false': {
            '()': 'django.utils.log.RequireDebugFalse'
        }
    },
    'formatters': {
        'simple': {
            'format': '%(levelname)s: %(message)s'
        },
        'verbose': {
            'format': '[%(asctime)s][%(module)s] %(levelname)s: %(message)s'
        }
    },
    'handlers': {
        'eoxserver_file': {
            'level': 'DEBUG',
            'class': 'logging.FileHandler',
            'filename': join(PROJECT_DIR, 'logs', 'eoxserver.log'),
            'formatter': 'verbose' if DEBUG else 'simple',
            'filters': [],
        },
        'stderr_stream': {
            'level': 'INFO',
            'class': 'logging.StreamHandler',
            'formatter': 'simple',
            'filters': [],
        },
    },
    'loggers': {
        'eoxserver': {
            'handlers': ['eoxserver_file'],
            'level': 'DEBUG' if DEBUG else 'INFO',
            'propagate': False,
        },
        'django': {
            'handlers': ['stderr_stream'],
            'level': 'WARNING',
            'propagate': False,
        },
    }
}

FIXTURE_DIRS = (
    join(PROJECT_DIR, 'data/fixtures'),
)

# Set this variable if the path to the instance cannot be resolved
# automatically, e.g. in case of redirects
#FORCE_SCRIPT_NAME="/path/to/instance/"



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

import contextlib
import time
import logging
try:
    from cgi import escape
except ImportError:
    from html import escape
import tempfile
import os

try:
    from mapscript import *
except ImportError:
    # defaults for document generation
    if os.environ.get('READTHEDOCS', None) == 'True':
        class mapObj: pass
        class layerObj: pass
        class classObj: pass
        class styleObj: pass
        class shapeObj: pass
        class colorObj: pass
        MS_LAYER_RASTER = 0
        MS_LAYER_POLYGON = 1
        MS_GET_REQUEST = 0
    else:
        raise
else:
    msversion = msGetVersionInt()

from lxml import etree
from eoxserver.core.util.multiparttools import iterate
from eoxserver.contrib import gdal
from eoxserver.services.result import result_set_from_raw_data

logger = logging.getLogger(__name__)


class MapServerException(Exception):
    def __init__(self, message, locator, code=None):
        super(MapServerException, self).__init__(message)
        self.locator = locator
        self.code = code


class MetadataMixIn(object):
    """ Mix-In for classes that wrap mapscript objects with associated metadata.
    """

    def __init__(self, metadata=None):
        super(MetadataMixIn, self).__init__()
        if metadata:
            self.setMetaData(metadata)

    def __getitem__(self, key):
        self.getMetaData(key)

    def __setitem__(self, key, value):
        self.setMetaData(key, value)

    def setMetaData(self, key_or_params, value=None, namespace=None):
        """ Convenvience method to allow setting multiple metadata values with
            one call and optionally setting a 'namespace' for each entry.
        """
        if value is None:
            for key, value in key_or_params.items():
                if namespace:
                    key = "%s_%s" % (namespace, key)

                super(MetadataMixIn, self).setMetaData(key, value)
        else:
            if namespace:
                key = "%s_%s" % (namespace, key_or_params)
            else:
                key = key_or_params

            return super(MetadataMixIn, self).setMetaData(key, value)


class Map(MetadataMixIn, mapObj):
    def dispatch(self, request):
        return dispatch(self, request)


def dispatch(map_, request):
    """ Wraps the ``OWSDispatch`` method. Perfoms all necessary steps for a
        further handling of the result.
    """

    logger.debug("MapServer: Installing stdout to buffer.")
    msIO_installStdoutToBuffer()

    # write the map if debug is enabled
    if logger.isEnabledFor(logging.DEBUG):
        fd, filename = tempfile.mkstemp(text=True)
        try:
            with os.fdopen(fd) as f:
                map_.save(filename)
                logger.debug(f.read())
        finally:
            os.remove(filename)

    try:
        logger.debug("MapServer: Dispatching.")
        ts = time.time()
        # Execute the OWS request by mapserver, obtain the status in
        # dispatch_status (0 is OK)
        status = map_.OWSDispatch(request)
        te = time.time()
        logger.debug("MapServer: Dispatch took %f seconds." % (te - ts))
    except Exception as e:
        raise MapServerException(str(e), "NoApplicableCode")

    raw_bytes = msIO_getStdoutBufferBytes()

    # check whether an error occurred
    if status != 0:
        # First try to get the error message through the error object
        obj = msGetErrorObj()
        if obj and obj.message:
            raise MapServerException(obj.message, obj.code)

        try:
            # try to parse the output as XML
            result = result_set_from_raw_data(raw_bytes)
            tree = etree.fromstring(result[0].data)
            exception_elem = tree.xpath("*[local-name() = 'Exception']|*[local-name() = 'ServiceException']")[0]
            locator = exception_elem.attrib.get("locator")
            code = exception_elem.attrib.get("exceptionCode")
            message = exception_elem[0].text

            raise MapServerException(message, locator, code)

        except (etree.XMLSyntaxError, IndexError, KeyError):
            pass

        # Fallback: raise arbitrary error
        raise MapServerException("Unexpected Error.", "NoApplicableCode")

    logger.debug("MapServer: Performing MapServer cleanup.")
    # Workaround for MapServer issue #4369
    if msversion < 60004 or (msversion < 60200 and msversion >= 60100):
        msCleanup()
    else:
        msIO_resetHandlers()

    return raw_bytes


class Layer(MetadataMixIn, layerObj):
    def __init__(self, name, metadata=None, type=MS_LAYER_RASTER, mapobj=None):
        layerObj.__init__(self, mapobj)
        MetadataMixIn.__init__(self, metadata)
        self.name = name
        self.status = MS_ON
        if type == MS_LAYER_RASTER:
            self.dump = MS_TRUE
            self.setConnectionType(MS_RASTER, "")
        self.type = type


class Class(classObj):
    def __init__(self, name, mapobj=None):
        classObj.__init__(self, mapobj)
        self.name = name


class Style(styleObj):
    def __init__(self, name, mapobj=None):
        styleObj.__init__(self, mapobj)
        self.name = name


def create_request(values, request_type=MS_GET_REQUEST):
    """ Creates a mapserver request from
    """
    used_keys = {}

    request = OWSRequest()
    request.type = request_type

    if request_type == MS_GET_REQUEST:
        for key, value in values:
            key = key.lower()
            used_keys.setdefault(key, 0)
            # addParameter() available in MapServer >= 6.2
            # https://github.com/mapserver/mapserver/issues/3973
            try:
                request.addParameter(key.lower(), escape(value))
            # Workaround for MapServer 6.0
            except AttributeError:
                used_keys[key] += 1
                request.setParameter(
                    "%s_%d" % (key, used_keys[key]), escape(value)
                )
    elif request_type == MS_POST_REQUEST:
        request.postrequest = value

    return request


def gdalconst_to_imagemode(const):
    """
    This function translates a GDAL data type constant as defined in the
    :mod:`gdalconst` module to a MapScript image mode constant.
    """
    if const == gdal.GDT_Byte:
        return MS_IMAGEMODE_BYTE
    elif const in (gdal.GDT_Int16, gdal.GDT_UInt16):
        return MS_IMAGEMODE_INT16
    elif const == gdal.GDT_Float32:
        return MS_IMAGEMODE_FLOAT32
    else:
        raise ValueError(
            "MapServer is not capable to process the datatype '%s' (%d)."
            % (gdal.GetDataTypeName(const), const)
        )


def gdalconst_to_imagemode_string(const):
    """
    This function translates a GDAL data type constant as defined in the
    :mod:`gdalconst` module to a string as used in the MapServer map file
    to denote an image mode.
    """
    if const == gdal.GDT_Byte:
        return "BYTE"
    elif const in (gdal.GDT_Int16, gdal.GDT_UInt16):
        return "INT16"
    elif const == gdal.GDT_Float32:
        return "FLOAT32"


def setMetaData(obj, key_or_params, value=None, namespace=None):
    """ Convenvience function to allow setting multiple metadata values with
        one call and optionally setting a 'namespace' for each entry.
    """
    if value is None:
        for key, value in key_or_params.items():
            if namespace:
                key = "%s_%s" % (namespace, key)

            obj.setMetaData(key, value)
    else:
        if namespace:
            key = "%s_%s" % (namespace, key_or_params)
        else:
            key = key_or_params

        obj.setMetaData(key, value)

# alias
set_metadata = setMetaData


def set_env(map_obj, env, fail_on_override=False, return_old=False):
    old_values = {} if return_old else None
    for key, value in env.items():
        if fail_on_override or return_old:
            old_value = map_obj.getConfigOption(str(key))
            if fail_on_override and old_value is not None \
                    and old_value != value:
                raise Exception(
                    'Would override previous value of %s: %s with %s'
                    % (key, old_value, value)
                )
            elif old_value != value and return_old:
                old_values[key] = old_value

        map_obj.setConfigOption(str(key), str(value))

    return old_values

@contextlib.contextmanager
def config_env(map_obj, env, fail_on_override=False, reset_old=True):
    old_env = set_env(env, fail_on_override, reset_old)
    yield
    if reset_old:
        set_env(old_env, False, False)



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Martin Paces <martin.paces@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2016 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------
# pylint: disable=missing-docstring

from optparse import make_option
from django.core.management.base import BaseCommand, CommandError
from eoxserver.resources.coverages.management.commands import (
    CommandOutputMixIn,
)
from eoxserver.services.management.commands.wms_options_load import (
    insert_or_update_wms_opt,
)


class Command(CommandOutputMixIn, BaseCommand):

    option_list = BaseCommand.option_list + (
        make_option(
            "-i", "--identifier", "--coverage-id", dest="identifier",
            action="store", default=None,
            help="Mandatory coverage identifier."
        ),
        make_option(
            "-r", "--red", dest="red", action="store", default=None,
            help="Band index of the red channel."
        ),
        make_option(
            "-g", "--green", dest="green", action="store", default=None,
            help="Band index of the green channel."
        ),
        make_option(
            "-b", "--blue", dest="blue", action="store", default=None,
            help="Band index of the blue channel."
        ),
        make_option(
            "-a", "--alpha", dest="alpha", action="store", default=None,
            help="Band index of the alpha channel."
        ),
        make_option(
            "--grey", "--grey-scale", dest="red", action="store", default=None,
            help="Index of the band displayed as a grey-scale image."
        ),
        make_option(
            "--resampling", dest="resampling", action="store",
            default=None, choices=["NEAREST", "AVERAGE", "BILINEAR"],
            help="Set image resampling method."
        ),
        make_option(
            "--min", "--scale-min", dest="scale_min", action="store",
            default=None,
            help=(
                "Range scale minimum value. Use a comma separated list "
                "if each band has a different value."
            )
        ),
        make_option(
            "--max", "--scale-max", dest="scale_max", action="store",
            default=None,
            help=(
                "Range scale minimum value. Use a comma separated list "
                "if each band has a different value."
            )
        ),
        make_option(
            "--auto-scale", dest="scale_auto", action="store_true",
            default=False,
            help="Enable automatic range scaling."
        ),
        make_option(
            "--no-auto-scale", dest="scale_auto", action="store_false",
            help="Disable automatic range scaling."
        ),
    )

    args = "<coverage-id>"

    help = """
    Set WMS options for a coverage.
    
    NOTE: Band indices are counted from 1.
    """

    def handle(self, *args, **options):
        # Collect parameters
        self.traceback = bool(options.get("traceback", False))
        self.verbosity = int(options.get('verbosity', 1))

        try:
            coverage_id = args[0]
        except IndexError:
            raise CommandError("Missing the mandatory coverage identifier!")

        keys = set((
            'red', 'green', 'blue', 'alpha', 'resampling', 'scale_auto'
        ))
        wms_opts = dict(
            (key, val) for key, val in options.items() if key in keys
        )

        if options['scale_min']:
            if ',' in options['scale_min']:
                wms_opts['bands_scale_min'] = options['scale_min']
            else:
                wms_opts['scale_min'] = options['scale_min']

        if options['scale_max']:
            if ',' in options['scale_max']:
                wms_opts['bands_scale_max'] = options['scale_max']
            else:
                wms_opts['scale_max'] = options['scale_max']

        # write the options
        insert_or_update_wms_opt(coverage_id, wms_opts)



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from eoxserver.core.util.timetools import isoformat
from eoxserver.services.subset import Slice
from eoxserver.services.ows.wcs.parameters import (
    CoverageRenderParams, CoverageDescriptionRenderParams,
    WCSCapabilitiesRenderParams
)


class WCS20CapabilitiesRenderParams(WCSCapabilitiesRenderParams):
    def __init__(self, coverages, dataset_series=None, sections=None,
                 accept_languages=None, accept_formats=None,
                 updatesequence=None, request=None):
        super(WCS20CapabilitiesRenderParams, self).__init__(
            coverages, "2.0.1", sections, accept_languages, accept_formats,
            updatesequence, request
        )
        self._dataset_series = dataset_series

    dataset_series = property(lambda self: self._dataset_series)


class WCS20CoverageDescriptionRenderParams(CoverageDescriptionRenderParams):
    coverage_ids_key_name = "coverageid"

    def __init__(self, coverages):
        super(WCS20CoverageDescriptionRenderParams, self).__init__(
            coverages, "2.0.1"
        )


class WCS20CoverageRenderParams(CoverageRenderParams):
    def __init__(self, coverage, subsets=None, rangesubset=None, format=None,
                 outputcrs=None, mediatype=None, interpolation=None,
                 scalefactor=None, scales=None, encoding_params=None,
                 http_request=None):

        super(WCS20CoverageRenderParams, self).__init__(coverage, "2.0.1")
        self._subsets = subsets
        self._rangesubset = rangesubset or ()
        self._scalefactor = scalefactor
        self._scales = scales or ()
        self._format = format
        self._outputcrs = outputcrs
        self._mediatype = mediatype
        self._interpolation = interpolation
        self._encoding_params = encoding_params or {}
        self._http_request = http_request


    coverage_id_key_name = "coverageid"

    subsets       = property(lambda self: self._subsets)
    rangesubset   = property(lambda self: self._rangesubset)
    scalefactor   = property(lambda self: self._scalefactor)
    scales        = property(lambda self: self._scales)
    format        = property(lambda self: self._format)
    outputcrs     = property(lambda self: self._outputcrs)
    mediatype     = property(lambda self: self._mediatype)
    interpolation = property(lambda self: self._interpolation)
    encoding_params = property(lambda self: self._encoding_params)
    http_request  = property(lambda self: self._http_request)


    def __iter__(self):
        for k, v in super(WCS20CoverageRenderParams, self).__iter__():
            yield k, v

        for subset in self.subsets:
            yield self.subset_to_kvp(subset)

        if self.format:
            yield ("format", self.format)

        if self.outputcrs:
            yield ("outputcrs", self.outputcrs)

        if self.mediatype:
            yield ("mediatype", self.mediatype)

        if self.interpolation:
            yield ("interpolation", self.interpolation)


    def subset_to_kvp(self, subset):
        temporal_format = lambda v: ('"%s"' % isoformat(v) if v else "*")
        spatial_format = lambda v: (str(v) if v is not None else "*")

        frmt = temporal_format if subset.is_temporal else spatial_format

        if isinstance(subset, Slice):
            value = frmt(subset.value)
        else:
            value = "%s,%s" % (frmt(subset.low), frmt(subset.high))

        crs = self.subsets.crs
        if crs:
            return "subset", "%s,%s(%s)" % (subset.axis, crs, value)
        else:
            return "subset", "%s(%s)" % (subset.axis, value)



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from eoxserver.core import Component, implements, UniqueExtensionPoint
from eoxserver.core.decoders import kvp, InvalidParameterException
from eoxserver.resources.coverages import models
from eoxserver.services.ows.interfaces import (
    ServiceHandlerInterface, GetServiceHandlerInterface
)
from eoxserver.services.ows.wms.util import LayerSelection
from eoxserver.services.ows.wms.interfaces import (
    WMSLegendGraphicRendererInterface
)
from eoxserver.services.result import to_http_response


class WMS13GetLegendGraphicHandler(Component):
    implements(ServiceHandlerInterface)
    implements(GetServiceHandlerInterface)

    renderer = UniqueExtensionPoint(WMSLegendGraphicRendererInterface)

    service = "WMS"
    versions = ("1.3.0", "1.3")
    request = "GetLegendGraphic"

    def handle(self, request):
        decoder = WMS13GetLegendGraphicDecoder(request.GET)

        layer_name = decoder.layer
        coverage_id = decoder.coverage

        suffixes = self.renderer.suffixes
        for suffix in suffixes:
            try:
                if len(suffix or "") == 0:
                    identifier = layer_name
                else:
                    identifier = layer_name[-len(suffix):]
                eo_object = models.EOObject.objects.get(identifier=identifier)
                break
            except models.EOObject.DoesNotExist:
                pass
        else:
            raise InvalidParameterException(
                "No such layer '%s'." % layer_name, "layer"
            )

        if models.iscollection(eo_object):
            def recursive_lookup(collection, used_ids, suffix):
                eo_objects = models.EOObject.objects.filter(
                    collections__in=[collection.pk]
                ).exclude(
                    pk__in=used_ids
                )

                result = []
                for eo_object in eo_objects:
                    used_ids.add(eo_object.pk)

                    if models.iscoverage(eo_object):
                        result.append((eo_object.cast(), suffix))
                    elif models.iscollection(eo_object):
                        result.extend(
                            recursive_lookup(eo_object, used_ids, suffix)
                        )
                    else:
                        pass

                return result

            used_ids = set()
            coverages = recursive_lookup(eo_object, used_ids, suffix)
            collection = eo_object

            if coverage_id:
                for coverage in coverages:
                    if coverage.identifier == coverage_id:
                        coverages = ((coverage, suffix),)
                        break
                else:
                    raise InvalidParameterException(
                        "Layer '%s' does not contain a coverage with ID '%s'.",
                        "coverage"
                    )
        else:
            collection = None
            coverages = ((eo_object.cast(), suffix),)

        layer_selection = LayerSelection(
            collection if collection else None
        )
        layer_selection.extend(coverages)

        result, _ = self.renderer.render(layer_selection, request.GET.items())
        return to_http_response(result)


class WMS13GetLegendGraphicDecoder(kvp.Decoder):
    layer    = kvp.Parameter(num=1)
    coverage = kvp.Parameter(num="?")



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2017 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from django.core.management.base import CommandError, BaseCommand
from django.db import transaction
from django.db.models import Q

from eoxserver.resources.coverages import models
from eoxserver.resources.coverages.management.commands import (
    CommandOutputMixIn, SubParserMixIn
)


class Command(CommandOutputMixIn, SubParserMixIn, BaseCommand):
    """ Command to manage mosaics. This command uses sub-commands for the
        specific tasks: create, delete, insert, exclude, purge.
    """
    def add_arguments(self, parser):
        create_parser = self.add_subparser(parser, 'create')
        delete_parser = self.add_subparser(parser, 'delete')
        insert_parser = self.add_subparser(parser, 'insert')
        exclude_parser = self.add_subparser(parser, 'exclude')
        refresh_parser = self.add_subparser(parser, 'refresh')
        purge_parser = self.add_subparser(parser, 'purge')
        parsers = [
            create_parser, delete_parser, insert_parser, exclude_parser,
            refresh_parser, purge_parser
        ]

        # identifier is a common argument
        for parser in parsers:
            parser.add_argument(
                'identifier', nargs=1, help='The mosaic identifier'
            )

        create_parser.add_argument(
            '--type', '-t', dest='type_name', required=True,
            help='The coverage type name of the mosaic. Mandatory.'
        )
        create_parser.add_argument(
            '--grid', '-g', dest='grid_name', default=None,
            help='The optional grid name.'
        )

        # common arguments for insertion/exclusion
        insert_parser.add_argument(
            'coverage_identifiers', nargs='+',
            help='The identifiers of the coverages to insert'
        )
        exclude_parser.add_argument(
            'coverage_identifiers', nargs='+',
            help=(
                'The identifiers of the coverages to exclude'
            )
        )

    @transaction.atomic
    def handle(self, subcommand, identifier, *args, **kwargs):
        """ Dispatch sub-commands: create, delete, insert, exclude, purge.
        """
        identifier = identifier[0]
        if subcommand == "create":
            self.handle_create(identifier, *args, **kwargs)
        elif subcommand == "delete":
            self.handle_delete(identifier, *args, **kwargs)
        elif subcommand == "insert":
            self.handle_insert(identifier, *args, **kwargs)
        elif subcommand == "exclude":
            self.handle_exclude(identifier, *args, **kwargs)
        elif subcommand == "refresh":
            self.handle_refresh(identifier, *args, **kwargs)
        elif subcommand == "purge":
            self.handle_purge(identifier, *args, **kwargs)

    def handle_create(self, identifier, type_name, grid_name, **kwargs):
        """ Handle the creation of a new mosaic.
        """
        if grid_name:
            try:
                grid = models.Grid.objects.get(name=grid_name)
            except models.Grid.DoesNotExist:
                raise CommandError("Grid %r does not exist." % grid_name)
        else:
            grid = None

        try:
            coverage_type = models.CoverageType.objects.get(
                name=type_name
            )
        except models.CoverageType.DoesNotExist:
            raise CommandError(
                "Coverage type %r does not exist." % type_name
            )

        models.Mosaic.objects.create(
            identifier=identifier,
            coverage_type=coverage_type, grid=grid,
            axis_1_size=0,
        )

        self.print_msg(
            'Successfully inserted created mosaic %s' % identifier
        )

    def handle_delete(self, identifier, **kwargs):
        """ Handle the deletion of a mosaic
        """
        mosaic = self.get_mosaic(identifier)
        grid = mosaic.grid
        mosaic.delete()

        grid_used = models.EOObject.objects.filter(
            Q(coverage__grid=grid) | Q(mosaic__grid=grid),
        ).exists()

        # clean up grid as well, if it is not referenced anymore
        # but saving named (user defined) grids
        if grid and not grid.name and not grid_used:
            grid.delete()

        self.print_msg(
            'Successfully deleted mosaic %s' % identifier
        )

    def handle_insert(self, identifier, coverage_identifiers, **kwargs):
        """ Handle the insertion of coverages into a mosaic
        """
        mosaic = self.get_mosaic(identifier)

        coverages = list(
            models.Coverage.objects.filter(
                identifier__in=coverage_identifiers
            )
        )

        if len(coverages) != len(set(coverage_identifiers)):
            actual = set(obj.identifier for obj in coverages)
            missing = set(coverage_identifiers) - actual
            raise CommandError(
                "No such coverage with ID%s: %s"
                % ("s" if len(missing) > 1 else "", ", ".join(missing))
            )

        for coverage in coverages:
            try:
                models.mosaic_insert_coverage(mosaic, coverage)
                self.print_msg(
                    'Successfully inserted coverage %s into mosaic %s'
                    % (coverage.identifier, mosaic.identifier)
                )
            except Exception as e:
                raise CommandError(
                    "Could not insert coverage %r into mosaic %r. "
                    "Error was: %s"
                    % (coverage.identifier, mosaic.identifier, e)
                )

    def handle_exclude(self, identifier, coverage_identifiers, **kwargs):
        """ Handle the exclusion of arbitrary objects from a mosaic
        """
        mosaic = self.get_mosaic(identifier)

        coverages = list(
            models.Coverage.objects.filter(
                identifier__in=coverage_identifiers
            )
        )

        if len(coverages) != len(set(coverage_identifiers)):
            actual = set(obj.identifier for obj in coverages)
            missing = set(coverage_identifiers) - actual
            raise CommandError(
                "No such object with ID%s: %s"
                % (len(missing) > 1, ", ".join(missing))
            )

        for coverage in coverages:
            try:
                models.mosaic_exclude_coverage(mosaic, coverage)
                self.print_msg(
                    'Successfully excluded coverage %s from mosaic %s'
                    % (coverage.identifier, mosaic.identifier)
                )
            except Exception as e:
                raise CommandError(
                    "Could not exclude coverage %r from mosic %r. "
                    "Error was: %s"
                    % (coverage.identifier, mosaic.identifier, e)
                )

    def handle_refresh(self, identifier, **kwargs):
        mosaic = self.get_mosaic(identifier)
        try:
            models.mosaic_recalc_metadata(mosaic)
            mosaic.full_clean()
            mosaic.save()

            self.print_msg(
                'Successfully refreshed metadata for mosaic %s'
                % (mosaic.identifier)
            )
        except Exception as e:
            raise CommandError(
                "Could not refresh metadata for mosaic %r. "
                "Error was: %s"
                % (mosaic.identifier, e)
            )

    def handle_purge(self, identifier, **kwargs):
        pass

    def get_mosaic(self, identifier):
        """ Helper method to get a mosaic by identifier or raise a
            CommandError.
        """
        try:
            return models.Mosaic.objects.get(identifier=identifier)
        except models.Mosaic.DoesNotExist:
            raise CommandError("Mosaic %r does not exist." % identifier)



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


from eoxserver.core import Component, implements
from eoxserver.core.decoders import xml, kvp, typelist
from eoxserver.services.ows.interfaces import (
    ServiceHandlerInterface, GetServiceHandlerInterface, 
    PostServiceHandlerInterface
)
from eoxserver.services.ows.wcs.basehandlers import WCSGetCoverageHandlerBase
from eoxserver.services.ows.wcs.v11.util import nsmap
from eoxserver.services.ows.wcs.v11.parameters import WCS11CoverageRenderParams


class WCS11GetCoverageHandler(WCSGetCoverageHandlerBase, Component):
    implements(ServiceHandlerInterface)
    implements(GetServiceHandlerInterface)
    implements(PostServiceHandlerInterface)

    versions = ("1.1.0", "1.1.1", "1.1.2")
    methods = ['GET', 'POST']

    def get_decoder(self, request):
        if request.method == "GET":
            return WCS11GetCoverageKVPDecoder(request.GET)
        elif request.method == "POST":
            return WCS11GetCoverageXMLDecoder(request.body)


    def get_params(self, coverage, decoder, request):
        return WCS11CoverageRenderParams(
            coverage, decoder.boundingbox, decoder.format, decoder.gridcs, 
            decoder.gridbasecrs, decoder.gridtype, decoder.gridorigin, 
            decoder.gridoffsets
        )


def parse_bbox_kvp(string):
    minx, miny, maxx, maxy, crs = string.split(",")
    minx, miny, maxx, maxy = map(float, (minx, miny, maxx, maxy))
    return minx, miny, maxx, maxy, crs


def parse_bbox_xml(node):
    try:
        lower_corner = node.xpath("ows:LowerCorner/text()", namespaces=nsmap)[0]
        upper_corner = node.xpath("ows:UpperCorner/text()", namespaces=nsmap)[0]
        ll = list(map(float, lower_corner.split(" ")))
        ur = list(map(float, upper_corner.split(" ")))
    except (IndexError, ValueError):
        raise ValueError("Invalid bounding box.")
    crs = node.attrib["crs"]
    return ll[0], ll[1], ur[0], ur[1], crs


def parse_origin_kvp(string):
    x, y = map(float, string.split(","))
    return x, y


def parse_origin_xml(string):
    x, y = map(float, string.split(" "))
    return x, y


def parse_offsets_kvp(string):
    x, y = map(float, string.split(","))
    return x, y


def parse_offsets_xml(string):
    x, y = map(float, string.split(" "))
    return x, y


class WCS11GetCoverageKVPDecoder(kvp.Decoder):
    coverage_id = kvp.Parameter("identifier", num=1)
    boundingbox = kvp.Parameter(type=parse_bbox_kvp, num=1)
    format = kvp.Parameter(num=1)
    gridcs = kvp.Parameter(num="?")
    gridbasecrs = kvp.Parameter(num="?")
    gridtype = kvp.Parameter(num="?")
    gridorigin = kvp.Parameter(type=parse_origin_kvp, num="?")
    gridoffsets = kvp.Parameter(type=parse_offsets_kvp, num="?")


class WCS11GetCoverageXMLDecoder(xml.Decoder):
    coverage_id = xml.Parameter("ows:Identifier/text()", type=str, num=1)
    boundingbox = xml.Parameter("wcs:DomainSubset/ows:BoundingBox", type=parse_bbox_xml, num=1)
    format = xml.Parameter("wcs:Output/@format", num=1)
    gridcs = xml.Parameter("wcs:Output/wcs:GridCRS/wcs:GridCS/text()", num="?")
    gridbasecrs = xml.Parameter("wcs:Output/wcs:GridCRS/wcs:GridBaseCRS/text()", num="?")
    gridtype = xml.Parameter("wcs:Output/wcs:GridCRS/wcs:GridType/text()", num="?")
    gridorigin = xml.Parameter("wcs:Output/wcs:GridCRS/wcs:GridOrigin/text()", type=parse_origin_xml, num="?")
    gridoffsets = xml.Parameter("wcs:Output/wcs:GridCRS/wcs:GridOffsets/text()", type=parse_offsets_xml, num="?")

    # TODO
    #interpolation = xml.Parameter("wcs:RangeSubset/wcs:InterpolationType/text()", num="?")
    #fields = xml.Parameter("wcs:RangeSubset/ows:Identifier/text()", num="*")

    namespaces = nsmap



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


import itertools

from eoxserver.core import env, Component, ExtensionPoint
from eoxserver.backends.interfaces import *


class BackendComponent(Component):
    """ This :class:`Component <eoxserver.core.component.Component>` provides
    extension points and helpers to easily retrive Package and Storage
    components by their type names.
    """

    file_storages = ExtensionPoint(FileStorageInterface)
    connected_storages = ExtensionPoint(ConnectedStorageInterface)
    packages = ExtensionPoint(PackageInterface)

    @property
    def storages(self):
        """ Helper to retrieve components for all storage interfaces.
        """
        return itertools.chain(self.file_storages, self.connected_storages)

    def get_file_storage_component(self, storage_type):
        """ Retrieve a component implementing the
        :class:`eoxserver.backends.interfaces.FileStorageInterface` with the
        desired ``storage_type``.

        :param storage_type: the desired storage type
        :returns: the desired storage component or ``None``
        """

        storage_type = storage_type.upper()
        result_component = None
        for storage_component in self.file_storages:
            if storage_component.name.upper() == storage_type:
                if result_component is not None:
                    raise Exception("Ambigouus storage component")
                result_component = storage_component

        return result_component

    def get_connected_storage_component(self, storage_type):
        """ Retrieve a component implementing the
        :class:`eoxserver.backends.interfaces.ConnectedStorageInterface` with
        the desired ``storage_type``.

        :param storage_type: the desired storage type
        :returns: the desired storage component or ``None``
        """

        storage_type = storage_type.upper()
        result_component = None
        for storage_component in self.connected_storages:
            if storage_component.name.upper() == storage_type:
                if result_component is not None:
                    raise Exception("Ambigouus storage component")
                result_component = storage_component

        return result_component

    def get_storage_component(self, storage_type):
        """ Retrieve a component implementing the
        :class:`eoxserver.backends.interfaces.FileStorageInterface` or
        :class:`eoxserver.backends.interfaces.ConnectedStorageInterface` with
        the desired ``storage_type``.

        :param storage_type: the desired storage type
        :returns: the desired storage component or ``None``
        """

        file_storage = self.get_file_storage_component(storage_type)
        connected_storage = self.get_connected_storage_component(storage_type)

        if file_storage is not None and connected_storage is not None:
            raise Exception("Ambigouus storage component")

        return file_storage or connected_storage

    def get_package_component(self, format):
        """ Retrieve a component implementing the
        :class:`eoxserver.backends.interfaces.PackageInterface` with
        the desired ``format``.

        :param format: the desired package format
        :returns: the desired package component or ``None``
        """

        format = format.upper()
        result_component = None
        for package_component in self.packages:
            if package_component.name.upper() == format:
                if result_component is not None:
                    raise Exception("Ambigouus package component")
                result_component = package_component

        return result_component


BackendComponent(env)



#-------------------------------------------------------------------------------
#
#  WPS Literal Data - allowed values
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Martin Paces <martin.paces@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2014 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from math import isinf
try:
    from itertools import chain, ifilterfalse as filterfalse
except ImportError:
    from itertools import filterfalse

from .data_types import BaseType, Double, DTYPES


class TypedMixIn(object):
    """ Mix-in class adding date-type to an allowed value range. """
    # pylint: disable=too-few-public-methods

    def __init__(self, dtype):
        if issubclass(dtype, BaseType):
            self._dtype = dtype
        elif dtype in DTYPES:
            self._dtype = DTYPES[dtype]
        else:
            raise TypeError("Non-supported data type %s!" % dtype)

    @property
    def dtype(self):
        """ Get data-type. """
        return self._dtype


class BaseAllowed(object):
    """ Allowed values base class. """

    def check(self, value):
        """ check validity """
        raise NotImplementedError

    def verify(self, value):
        """ Verify the value."""
        raise NotImplementedError


class AllowedAny(BaseAllowed):
    """ Allowed values class allowing any value. """

    def check(self, value):
        return True

    def verify(self, value):
        return value


class AllowedByReference(BaseAllowed):
    """ Allowed values class defined by a reference.

    NOTE: As it is not how such a reference definition looks like
    this class has the same behaviour as the AllowedAny class.
    """
    # TODO: Implement proper handling of the allowed values defined by a reference.

    def __init__(self, url):
        self._url = url

    @property
    def url(self):
        """ Get the URL of the reference. """
        return self._url

    def check(self, value):
        return True

    def verify(self, value):
        return value


class AllowedEnum(BaseAllowed, TypedMixIn):
    """ Allowed values class allowing values from an enumerated set. """

    @staticmethod
    def _unique(values):
        """ Get list of order-preserved unique values and the corresponding
        unordered set.
        """
        vset = set()
        vlist = []
        vset_add = vset.add
        vlist_append = vlist.append

        for value in filterfalse(vset.__contains__, values):
            vset_add(value)
            vlist_append(value)
        return vlist, vset

    def __init__(self, values, dtype=Double):
        TypedMixIn.__init__(self, dtype)
        self._values_list, self._values_set = self._unique(
            self._dtype.parse(value) for value in values
        )

    @property
    def values(self):
        """ Get the allowed values. """
        return self._values_list

    def check(self, value):
        return self._dtype.parse(value) in self._values_set

    def verify(self, value):
        if self.check(value):
            return value
        raise ValueError("The value is not in the set of the allowed values.")


class AllowedRange(BaseAllowed, TypedMixIn):
    """ Allowed values class allowing values from a range.

    Constructor parameters:
                minval      range lower bound - set to None if unbound
                maxval      range upper bound - set to None if unbound
                closure    *'closed'|'open'|'open-closed'|'closed-open'
                spacing     uniform spacing of discretely sampled ranges
                spacing_rtol relative tolerance of the spacing match
        """

    ALLOWED_CLOSURES = ['closed', 'open', 'open-closed', 'closed-open']

    # NOTE: Use of spacing with float discrete range is not recommended.
    def __init__(self, minval, maxval, closure='closed',
                 spacing=None, spacing_rtol=1e-9, dtype=Double):
        # pylint: disable=too-many-arguments
        TypedMixIn.__init__(self, dtype)

        if not self._dtype.comparable:
            raise ValueError("Non-supported range data type '%s'!" % self._dtype)

        if closure not in self.ALLOWED_CLOSURES:
            raise ValueError("Invalid closure specification!")

        if minval is None and maxval is None:
            raise ValueError("Invalid range bounds!")

        if spacing_rtol < 0.0 or spacing_rtol > 1.0:
            raise ValueError("Invalid spacing relative tolerance!")

        self._closure = self.ALLOWED_CLOSURES.index(closure)
        self._minval = None if minval is None else self._dtype.parse(minval)
        self._maxval = None if maxval is None else self._dtype.parse(maxval)
        self._base = self._maxval if self._minval is None else self._minval

        # verify the spacing
        if spacing is not None:
            ddtype = self._dtype.get_diff_dtype()

            # check whether the type has difference operation defined
            if ddtype is None or ddtype.zero is None:
                raise TypeError(
                    "Spacing is not applicable for type '%s'!" % dtype
                )
            spacing = ddtype.parse(spacing)
            if spacing <= ddtype.zero:
                raise ValueError("Invalid spacing '%s'!" % spacing)

        self._spacing = spacing
        self._rtol = spacing_rtol

    @property
    def minval(self):
        """ Get the lower bound of the range. """
        return self._minval

    @property
    def maxval(self):
        """ Get the upper bound of the range. """
        return self._maxval

    @property
    def spacing(self):
        """ Get the range spacing. """
        return self._spacing

    @property
    def closure(self):
        """ Get the range closure type. """
        return self.ALLOWED_CLOSURES[self._closure]

    def _out_of_spacing(self, value):
        if self._spacing is None:
            return False
        ddtype = self._dtype.get_diff_dtype()
        tmp0 = ddtype.as_number(self._dtype.sub(value, self._base))
        tmp1 = ddtype.as_number(self.spacing)
        tmp2 = float(tmp0) / float(tmp1)
        if isinf(tmp2):
            return True
        return not self._rtol >= abs(tmp2 - round(tmp2))

    def _out_of_bounds(self, value):
        if value != value:  # simple type-safe NaN check (works in Python > 2.5)
            return True
        below = self._minval is not None and (
            value < self._minval or (
                value == self._minval and self._closure in (1, 2)
            )
        )
        above = self._maxval is not None and (
            value > self._maxval or (
                value == self._maxval and self._closure in (1, 3)
            )
        )
        return below or above

    def check(self, value):
        value = self._dtype.parse(value)
        return not (self._out_of_bounds(value) or self._out_of_spacing(value))

    def verify(self, value):
        parsed_value = self._dtype.parse(value)
        if self._out_of_bounds(parsed_value):
            raise ValueError("The value is not within the allowed range.")
        if self._out_of_spacing(parsed_value):
            raise ValueError("The value does not fit the requested spacing.")
        return value


class AllowedRangeCollection(BaseAllowed, TypedMixIn):
    """ Allowed value class allowing values from a collection of AllowedEnum
    and AllowedRange instances.
    """

    def __init__(self, *objs):
        if not objs:
            raise ValueError(
                "At least one AllowedEnum or AllowedRange object "
                "must be provided!"
            )
        TypedMixIn.__init__(self, objs[0].dtype)

        values = set()
        ranges = []

        for obj in objs:
            # Merge enumerated values into one set.
            if isinstance(obj, AllowedEnum):
                values.update(obj.values)
            # Collect ranges.
            elif isinstance(obj, AllowedRange):
                ranges.append(obj)
            else:
                raise ValueError(
                    "An object which is neither AllowedEnum"
                    " nor AllowedRange instance! OBJ=%r" % obj
                )

            # Check that all ranges and value sets are of the same type.
            if self.dtype != obj.dtype:
                raise TypeError("Data type mismatch!")

        self._enum = AllowedEnum(values, dtype=self.dtype)
        self._ranges = ranges

    @property
    def enum(self):
        """ Get merged set of the enumerated allowed values. """
        return self._enum

    @property
    def ranges(self):
        """ Get list of the allowed values' ranges. """
        return self._ranges

    def check(self, value):
        for obj in chain([self._enum], self._ranges):
            if obj.check(value):
                return True
        return False

    def verify(self, value):
        if self.check(value):
            return value
        raise ValueError(
            "The value does not match the range of the allowed values!"
        )



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2017 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------


DEFAULT_EOXS_COVERAGE_METADATA_FORMAT_READERS = [
    'eoxserver.resources.coverages.metadata.coverage_formats.gsc.GSCFormatReader',
    'eoxserver.resources.coverages.metadata.coverage_formats.dimap_general.DimapGeneralFormatReader',
    'eoxserver.resources.coverages.metadata.coverage_formats.eoom.EOOMFormatReader',
    'eoxserver.resources.coverages.metadata.coverage_formats.gdal_dataset.GDALDatasetMetadataReader',
    'eoxserver.resources.coverages.metadata.coverage_formats.inspire.InspireFormatReader',
    'eoxserver.resources.coverages.metadata.coverage_formats.native.NativeFormat',
    'eoxserver.resources.coverages.metadata.coverage_formats.native_config.NativeConfigFormatReader',
    'eoxserver.resources.coverages.metadata.coverage_formats.landsat8_l1.Landsat8L1CoverageMetadataReader',
]

DEFAULT_EOXS_COVERAGE_METADATA_GDAL_DATASET_FORMAT_READERS = [
    'eoxserver.resources.coverages.metadata.coverage_formats.gdal_dataset_envisat.GDALDatasetEnvisatMetadataFormatReader',
]

DEFAULT_EOXS_PRODUCT_METADATA_FORMAT_READERS = [
    'eoxserver.resources.coverages.metadata.product_formats.sentinel1.S1ProductFormatReader',
    'eoxserver.resources.coverages.metadata.product_formats.sentinel2.S2ProductFormatReader',
    'eoxserver.resources.coverages.metadata.product_formats.landsat8_l1.Landsat8L1ProductMetadataReader',
    'eoxserver.resources.coverages.metadata.coverage_formats.eoom.EOOMFormatReader',
    'eoxserver.resources.coverages.metadata.product_formats.gsc.GSCProductMetadataReader',
]



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


class OWS20ExceptionHandler(object):
    """ A Fallback exception handler. This class does on purpose not implement
        the ExceptionHandlerInterface and must be instantiated manually.
    """

    def handle_exception(self, request, exception):
        message = str(exception)
        version = "2.0.0"
        # NOTE: An existing 'code' attribute can be set to None value!!!
        code = getattr(exception, "code", None) or "NoApplicableCode"
        locator = getattr(exception, "locator", None)
        status_code = 400

        from eoxserver.services.ows.common.v20.encoders import (
            OWS20ExceptionXMLEncoder
        )
        encoder = OWS20ExceptionXMLEncoder()

        return (
            encoder.serialize(
                encoder.encode_exception(message, version, code, locator)
            ),
            encoder.content_type,
            status_code
        )



#-----------------------------------------------------------------------
#
# Description: 
#
# This file contains definition of WCS 1.1.x Transaction extension operation handler. 
#
#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Martin Paces <martin.paces@iguassu.cz>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 Iguassu Software Systems, a.s 
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

import os.path

import logging
import traceback 

#from eoxserver.core.util.xmltools import DOMElementToXML
#from eoxserver.core.exceptions import InternalError
from eoxserver.services.interfaces import OperationHandlerInterface
from eoxserver.services.interfaces import VersionHandlerInterface
from eoxserver.services.owscommon import OWSCommonVersionHandler

from eoxserver.core.system import System
from eoxserver.services.requests import Response
#from eoxserver.services.ogc import OGCExceptionHandler
from eoxserver.services.exceptions import InvalidRequestException
from eoxserver.services.base import BaseRequestHandler 

from eoxserver.services.ows.wcst.wcstXML import parseCoverageXML 
from eoxserver.services.ows.wcst.wcst11Transaction import wcst11Transaction
from eoxserver.services.ows.wcst.wcst11Context import contextCreate, contextDiscardSuccess, contextDiscardFailure, contextDiscardAsync
from eoxserver.services.ows.wcst.wcst11Exception import createXML_OWS11Exception, ExOperationNotSupported
# 
# NOTE: WCS-T allows service to be '1.1' only. Since the EOxServer accept version '1.1.0' only 
#       we need to define a new version handler dedicated to WCS-T only. 
#       If you (EOX guys) know a better solution please let me know.
#


logger = logging.getLogger(__name__)

class WCS11VersionHandler(OWSCommonVersionHandler):
    SERVICE = "wcs"
    
    REGISTRY_CONF = {
        "name": "WCS 1.1 Version Handler",
        "impl_id": "services.ows.wcs11Transaction.WCS11VersionHandler",
        "registry_values": {
            "services.interfaces.service": "wcs",
            "services.interfaces.version": "1.1"
        }
    }

WCS11VersionHandlerImplementation = VersionHandlerInterface.implement(WCS11VersionHandler)

#
# WCS-T operation handler 
#
NS_WCST11="http://www.opengis.net/wcs/1.1/wcst"

class WCS11TransactionHandler(BaseRequestHandler):

    REGISTRY_CONF = {
        "name": "WCS 1.1.x Transaction Handler",
        "impl_id": "services.ows.wcs11Transaction.WCS11TransactionHandler",
        "registry_values": {
            "services.interfaces.service": "wcs",
            "services.interfaces.version": "1.1",
            "services.interfaces.operation": "transaction"
        }
    }

    PARAM_SCHEMA = {
        "requestId": {"xml_location": "/{%s}RequestId"%NS_WCST11, "xml_type": "string", "kvp_key": "requestId", "kvp_type": "string"},
        "responseHandler": {"xml_location": "/{%s}ResponseHandler"%NS_WCST11, "xml_type": "string", "kvp_key": "responseHandler", "kvp_type": "string"},
        "actions": {"xml_location":"/{%s}InputCoverages/{%s}Coverage"%(NS_WCST11,NS_WCST11),"xml_type":"element[]","kvp_key":"coverages","kvp_type":"string"},
        } 

    def _handleException(self, req, exception):

        logger.debug("WCS110: Transaction() - Exception Handler!") 

        # dump exception to the log file 
        logger.error( traceback.format_exc() ) 

        return Response( content=createXML_OWS11Exception(exception) , content_type="text/xml" , status=500 ) 
        

    def _processRequest(self, req ):

        logger.debug("WCS110: Transaction()") 

        # KVP encoding is not supported by WCS-T 
        if req.http_req.method.upper() == "GET" : 
            raise ExOperationNotSupported("transaction","Operation does not support KVP encoding!")  

        # parse request 
        req.decoder.setSchema( self.PARAM_SCHEMA ) 

        # get actions (requests) and context (working env.)
        actions = map( parseCoverageXML , req.decoder.getValue('actions') ) 
        context = contextCreate( req.decoder.getValue('requestId') , req.decoder.getValue('responseHandler') )  

        try: 

            # execute transaction 
            response = wcst11Transaction( actions , context )

        except : #FAILURE 

            # discard permanent and temporary storage 
            contextDiscardFailure( context )
            raise 

        else : # SUCCESS 

            # clean-up TMP space of a synchronous operation 
            if not context['isAsync'] : 
                contextDiscardSuccess( context ) 
            # no TMP clean-up for asynnchronous operation 
            else : 
                contextDiscardAsync( context ) 

        # prepare the response 
        return Response( content=response , content_type="text/xml" , status=200 )  


WCS11TransactionHandlerImplementation = OperationHandlerInterface.implement(WCS11TransactionHandler)



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


""" This module provides Python file-object like access to VSI files.
"""


import os
from uuid import uuid4
from functools import wraps

if os.environ.get('READTHEDOCS', None) != 'True':
    import numpy

    from eoxserver.contrib.gdal import (
        VSIFOpenL, VSIFCloseL, VSIFReadL, VSIFWriteL, VSIFSeekL, VSIFTellL,
        VSIStatL, VSIFTruncateL, Unlink, Rename, FileFromMemBuffer
    )

    UINT32_MAX = numpy.iinfo('uint32').max

    rename = Rename

    unlink = remove = Unlink


def open(filename, mode="r"):
    """ A function mimicking the builtin function
    :func:`open <__builtins__.open>` but returning a :class:`VSIFile` instead.

    :param filename: the path to the file; this might also be any VSI special
                     path like "/vsicurl/..." or "/vsizip/...". See the `GDAL
                     documentation
                     <http://trac.osgeo.org/gdal/wiki/UserDocs/ReadInZip>`_
                     for reference.
    :param mode: the file opening mode
    :returns: a :class:`VSIFile`
    """
    return VSIFile(filename, mode)


def _ensure_open(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if self._handle is None:
            raise ValueError('I/O operation on closed file')
        return func(self, *args, **kwargs)
    return wrapper


class VSIFile(object):
    """ File-like object interface for VSI file API.

    :param filename: the path to the file; this might also be any VSI special
                     path like "/vsicurl/..." or "/vsizip/...". See the `GDAL
                     documentation
                     <http://trac.osgeo.org/gdal/wiki/UserDocs/ReadInZip>`_
                     and `manuals
                     <http://www.gdal.org/gdal_virtual_file_systems.html>`_
                     for reference.
    :param mode: the file opening mode
    """

    def __init__(self, filename, mode="r"):
        self._handle = VSIFOpenL(filename, mode)
        self._filename = filename

        if self._handle is None:
            raise IOError("Failed to open file '%s'." % self._filename)

    @property
    def name(self):
        """ Returns the filename referenced by this file
        """
        return self._filename

    @_ensure_open
    def read(self, size=UINT32_MAX):
        """ Read from the file. If no ``size`` is specified, read until the end
        of the file.

        :param size: the number of bytes to be read
        :returns: the bytes read as a string
        """

        value = VSIFReadL(1, size, self._handle)
        return value if value is not None else ''

    @_ensure_open
    def write(self, data):
        """ Write the buffer ``data`` to the file.

        :param data: the string buffer to be written
        """
        VSIFWriteL(data, 1, len(data), self._handle)

    @_ensure_open
    def tell(self):
        """ Return the current read/write offset of the file.

        :returns: an integer offset
        """
        return VSIFTellL(self._handle)

    @_ensure_open
    def seek(self, offset, whence=os.SEEK_SET):
        """ Set the new read/write offset in the file.

        :param offset: the new offset
        :param whence: how the offset shall be interpreted; possible options are
                       :const:`os.SEEK_SET`, :const:`os.SEEK_CUR` and
                       :const:`os.SEEK_END`
        """
        VSIFSeekL(self._handle, offset, whence)

    def close(self):
        """ Close the file.
        """
        if self._handle is not None:
            VSIFCloseL(self._handle)
        self._handle = None

    @property
    def closed(self):
        """ Return a boolean value to indicate whether or not the file is
        already closed.
        """
        return (self._handle is None)

    def __iter__(self):
        """ Iterate over the lines within the file.
        """
        return self

    @_ensure_open
    def next(self):
        """ Satisfaction of the iterator protocol. Return the next line in the
            file or raise `StopIteration`.
        """
        line = self.readline()
        if not line:
            raise StopIteration
        return line

    @_ensure_open
    def readline(self, length=None, windowsize=1024):
        """ Read a single line from the file and return it.

        :param length: the maximum number of bytes to read to look for a whole
                       line.
        :param windowsize: the windowsize to search for a newline character.
        """
        line = ""
        while True:
            # read amount and detect for EOF
            string = self.read(length or windowsize)
            if not string:
                break

            try:
                position = string.index('\n')
                line += string[:position]

                # retun the cursor for the remainder of the string
                self.seek(-(len(string) - (position + 1)), os.SEEK_CUR)
                break
            except ValueError:
                line += string

            # also break when a specific size was requested but no newline was
            # found
            if length:
                break

        return line

    def readlines(self, sizehint=0):
        """ Read the remainder of the file (or up to `sizehint` bytes) and return
            the lines.

        :param sizehint: the number of bytes to scan for lines.
        :return: the lines
        :rtype: list of strings
        """
        # TODO: take sizehint into account
        lines = [line for line in self]
        return lines

    @property
    @_ensure_open
    def size(self):
        """ Return the size of the file in bytes
        """
        stat = VSIStatL(self.name)
        return stat.size

    @_ensure_open
    def flush(self):
        pass
        # VSIFlushL(self._handle)  # TODO: not available?

    @_ensure_open
    def truncate(self, size=None):
        """ Truncates the file to the given size or to the size until the current
            position.

        :param size: the new size of the file.
        """
        size = size or self.tell()
        VSIFTruncateL(self._handle, size)

    def isatty(self):
        """ Never a TTY """
        return False

    def __enter__(self):
        return self

    def __exit__(self, etype=None, evalue=None, tb=None):
        self.close()

    def __del__(self):
        self.close()


class TemporaryVSIFile(VSIFile):
    """ Subclass of VSIFile, that automatically deletes the physical file upon
        deletion.
    """

    @classmethod
    def from_buffer(cls, buf, mode="wb", filename=None):
        """ Creates a :class:`TemporaryVSIFile` from a string.

        :param buf: the supplied string
        :param mode: the file opening mode
        :param filename: the optional filename the file shall be stored under;
                         by default this is an in-memory location
        """
        if not filename:
            filename = "/vsimem/%s" % uuid4().hex
        f = cls(filename, mode)
        f.write(buf)
        f.seek(0)
        return f

    def close(self):
        """ Close the file. This also deletes it.
        """
        if not self.closed:
            super(TemporaryVSIFile, self).close()
            remove(self.name)


def join(first, *paths):
    """ Joins the given VSI path specifiers. Similar to :func:`os.path.join` but
        takes care of the VSI-specific handles such as `vsicurl`, `vsizip`, etc.
    """
    parts = first.split('/')
    for path in paths:
        new = path.split('/')
        if path.startswith('/vsi'):
            parts = new[0:2] + (parts if parts[0] else parts[1:]) + new[2:]
        else:
            parts.extend(new)
    return '/'.join(parts)



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

import os
import hashlib
import logging
from fnmatch import fnmatch

from eoxserver.core.util.iteratortools import pairwise_iterative
from eoxserver.contrib import vsi, gdal
from eoxserver.backends.cache import get_cache_context
from eoxserver.backends.storages import get_handler_class_for_model
from eoxserver.backends import storage_auths


logger = logging.getLogger(__name__)


class AccessError(Exception):
    pass


def _generate_hash(location, format, hash_impl="sha1"):
    h = hashlib.new(hash_impl)
    if format is not None:
        h.update(format)
    h.update(location)
    return h.hexdigest()


def _linearize_storages(data_item):
    """ Retrieve a list of all storages.
    """
    chain = []
    storage = data_item.storage
    while storage:
        handler_cls = get_handler_class_for_model(storage)
        if not handler_cls:
            raise AccessError(
                'Unsupported storage type %r' % storage.storage_type
            )

        chain.append((storage, handler_cls))
        storage = storage.parent
    return reversed(chain)


def retrieve(data_item, cache=None):
    """ Retrieves the :class:`eoxserver.backends.models.DataItem` and makes the
        file locally available if necessary.
        When the data item is not associated with a storage, then the data items
        location will be returned. Otherwise, the storage  handlers ``retrieve``
        method will be called to make the data item locally available.

        :param data_item: data item to retrieve
        :type data_item: :class:`eoxserver.backends.models.DataItem`
        :param cache: the optional cache context
        :type cache: eoxserver.backends.cache.CacheContext
        :returns: the path to the localized file
        :rtype: str
    """
    cache = cache or get_cache_context()

    # use shortcut here, when no storage is provided
    if not data_item.storage:
        return data_item.location

    storage_handlers = _linearize_storages(data_item)
    with cache:
        handler = None
        path = None
        for current, child in pairwise_iterative(storage_handlers):
            storage, handler_cls = current
            child_storage, _ = child

            item_id = _generate_hash(data_item.location, data_item.format)
            tmp_path = cache.relative_path(item_id)
            if not cache.contains(item_id):
                # actually retrieve the item when not in the cache
                handler = handler_cls(path or storage.url)
                use_cache, path = handler.retrieve(
                    path or child_storage.url, tmp_path
                )
                if not use_cache:
                    cache.add_mapping(path)
            else:
                path = tmp_path

        if storage_handlers:
            storage, handler_cls = storage_handlers[-1]
            handler = handler_cls(path)
            return handler.retrieve(data_item.location)[1]


def open(data_item, cache=None):
    """ Returns a file object pointing to the given location. This function
    works like the builtin function :func:`open() <__builtins__.open>` but on
    a :class:`DataItem <eoxserver.backends.models.DataItem>` and performs a
    :func:`retrieve` first.

    :param data_item: the :class:`DataItem <eoxserver.backends.models.DataItem>`
                      to open
    :param cache: an instance of :class:`CacheContext
                  <eoxserver.backends.cache.CacheContext>` or ``None``
                  if the caching shall be handled internally
    """

    return __builtins__.open(retrieve(data_item, cache))


def get_vsi_path(data_item):
    """ Get the VSI path to the given :class:`eoxserver.backends.models.DataItem`

        :param data_item: the data item to get the path to
        :type data_item: :class:`eoxserver.backends.models.DataItem`
        :returns: the VSI file path which is can be used with GDAL-related APIs
        :rtype: str
    """
    location = data_item.location
    storage = data_item.storage

    return get_vsi_storage_path(storage, location)


def get_vsi_storage_path(storage, location=None):
    while storage:
        handler_cls = get_handler_class_for_model(storage)
        if handler_cls:
            handler = handler_cls(storage.url)
            location = handler.get_vsi_path(location or '')
        else:
            raise AccessError(
                'Unsupported storage type %r' % storage.storage_type
            )
        storage = storage.parent

    return location



def get_vsi_env(storage):
    """
    """
    env = {}
    while storage:
        handler_cls = get_handler_class_for_model(storage)
        if handler_cls:
            handler = handler_cls(storage.url)
            env.update(handler.get_vsi_env())
        else:
            raise AccessError(
                'Unsupported storage type %r' % storage.storage_type
            )

        if storage.storage_auth:
            handler = storage_auths.get_handler_for_model(
                storage.storage_auth
            )
            if handler:
                env.update(handler.get_vsi_env())
            else:
                raise AccessError(
                    'Unsupported storage auth type %r'
                    % storage.storage_auth.storage_auth_type
                )

        storage = storage.parent

    return env


def vsi_open(data_item):
    """ Opens a :class:`eoxserver.backends.models.DataItem` as a
        :class:`eoxserver.contrib.vsi.VSIFile`. Uses :func:`get_vsi_path`
        internally to get the path.

        :param data_item: the data item to open as a VSI file
        :type data_item: :class:`eoxserver.backends.models.DataItem`
        :rtype: :class:`eoxserver.contrib.vsi.VSIFile`
    """
    with gdal.config_env(get_vsi_env(data_item.storage)):
        return vsi.open(get_vsi_path(data_item))


def gdal_open(data_item, shared=True):
    """ Opens a :class:`eoxserver.backends.models.DataItem` as a
        :class:`eoxserver.contrib.gdal.Dataset`. Uses :func:`get_vsi_path`
        internally to get the path.

        :param data_item: the data item to open as a dataset.
        :type data_item: :class:`eoxserver.backends.models.DataItem`
        :rtype: :class:`eoxserver.contrib.gdal.Dataset`
    """
    return gdal.open_with_env(
        get_vsi_path(data_item),
        get_vsi_env(data_item.storage),
        shared
    )


def vsi_list_storage(storage, location=None, pattern=None):
    """
    """
    with gdal.config_env(get_vsi_env(storage)):
        path = get_vsi_storage_path(storage, location)
        files = gdal.ReadDir(path)

        if files is None:
            raise Exception('No such path "%s"' % path)

        if pattern:
            files = [
                filename
                for filename in files
                if fnmatch(filename, pattern)
            ]

        return files



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2017 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

import logging
import tempfile
from uuid import uuid4

from eoxserver.contrib import mapserver as ms
from eoxserver.contrib import vsi
from eoxserver.render.colors import BASE_COLORS, COLOR_SCALES
from eoxserver.render.mapserver.factories import get_layer_factories
from eoxserver.resources.coverages.formats import getFormatRegistry


logger = logging.getLogger(__name__)


# TODO: move this to render.map.exceptions
class MapRenderError(Exception):
    def __init__(self, message, code=None, locator=None):
        super(MapRenderError, self).__init__(message)
        self.code = code
        self.locator = locator


class MapserverMapRenderer(object):

    OUTPUTFORMATS = [
        ('')
    ]

    def get_geometry_styles(self):
        return BASE_COLORS.keys()

    def get_raster_styles(self):
        return COLOR_SCALES.keys()

    def get_supported_layer_types(self):
        layer_types = []
        for layer_factory in get_layer_factories():
            layer_types.extend(layer_factory.handled_layer_types)
        return set(layer_types)

    def get_supported_formats(self):
        return getFormatRegistry().getSupportedFormatsWMS()

    def render_map(self, render_map):
        # TODO: get layer creators for each layer type in the map
        map_obj = ms.mapObj()

        if render_map.bgcolor:
            map_obj.imagecolor.setHex("#" + render_map.bgcolor.lower())
        else:
            map_obj.imagecolor.setRGB(0, 0, 0)

        frmt = getFormatRegistry().getFormatByMIME(render_map.format)

        if not frmt:
            raise MapRenderError(
                'No such format %r' % render_map.format,
                code='InvalidFormat',
                locator='format'
            )

        outputformat_obj = ms.outputFormatObj(frmt.driver)

        outputformat_obj.transparent = (
            ms.MS_ON if render_map.transparent else ms.MS_OFF
        )
        outputformat_obj.mimetype = frmt.mimeType

        if frmt.defaultExt:
            if frmt.defaultExt.startswith('.'):
                extension = frmt.defaultExt[1:]
            else:
                extension = frmt.defaultExt

            outputformat_obj.extension = extension

        map_obj.setOutputFormat(outputformat_obj)

        #
        map_obj.setExtent(*render_map.bbox)
        map_obj.setSize(render_map.width, render_map.height)
        map_obj.setProjection(render_map.crs)
        map_obj.setConfigOption('MS_NONSQUARE', 'yes')

        layers_plus_factories = self._get_layers_plus_factories(render_map)
        layers_plus_factories_plus_data = [
            (layer, factory, factory.create(map_obj, layer))
            for layer, factory in layers_plus_factories
        ]

        # log the resulting map
        if logger.isEnabledFor(logging.DEBUG):
            with tempfile.NamedTemporaryFile() as f:
                map_obj.save(f.name)
                f.seek(0)
                logger.debug(f.read().decode('ascii'))

        try:
            # actually render the map
            image_obj = map_obj.draw()

            try:
                image_bytes = image_obj.getBytes()
            except Exception:
                tmp_name = '/vsimem/%s' % uuid4().hex
                image_obj.save(tmp_name, map_obj)
                with vsi.open(tmp_name) as f:
                    image_bytes = f.read()
                vsi.unlink(tmp_name)

            extension = outputformat_obj.extension
            if extension:
                if len(render_map.layers) == 1:
                    filename = '%s.%s' % (
                        render_map.layers[0].name, extension
                    )
                else:
                    filename = 'map.%s' % extension
            else:
                filename = None

            return image_bytes, outputformat_obj.mimetype, filename

        finally:
            # disconnect
            for layer, factory, data in layers_plus_factories_plus_data:
                factory.destroy(map_obj, layer, data)

    def _get_layers_plus_factories(self, render_map):
        layers_plus_factories = []
        type_to_layer_factory = {}
        for layer in render_map.layers:
            layer_type = type(layer)
            if layer_type in type_to_layer_factory:
                factory = type_to_layer_factory[layer_type]
            else:
                factory = self._get_layer_factory(layer_type)
                type_to_layer_factory[layer_type] = factory

            layers_plus_factories.append((layer, factory()))

        return layers_plus_factories

    def _get_layer_factory(self, layer_type):
        for factory in get_layer_factories():
            if factory.supports(layer_type):
                return factory
        raise MapRenderError(
            'Could not find a layer factory for %r' % layer_type.__name__
        )



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2017 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from uuid import uuid4
import ast
import _ast
import operator
import logging

import numpy as np
from django.utils.six import string_types

from eoxserver.render.browse.util import warp_fields
from eoxserver.contrib import vrt, gdal, osr


logger = logging.getLogger(__name__)


class BrowseGenerationError(Exception):
    pass


class BrowseGenerator(object):
    def __init__(self, footprint_alpha=True, ):
        pass

    def generate(self, product, browse_type, style, out_filename):
        if not product.product_type or \
                not product.product_type == browse_type.product_type:
            raise BrowseGenerationError("Product and browse type don't match")


class FilenameGenerator(object):
    """ Utility class to generate filenames after a certain pattern (template)
        and to keep a list for later cleanup.
    """
    def __init__(self, template, default_extension=None):
        """ Create a new :class:`FilenameGenerator` from a given template
            :param template: the template string used to construct the
                             filenames from. Uses the ``.format()`` style
                             language. Keys are ``index``, ``uuid`` and
                             ``extension``.
        """
        self._template = template
        self._filenames = []
        self._default_extension = default_extension

    def generate(self, extension=None):
        """ Generate and store a new filename using the specified template. An
            optional ``extension`` can be passed, when used in the template.
        """
        filename = self._template.format(
            index=len(self._filenames),
            uuid=uuid4().hex,
            extension=extension or self._default_extension,
        )
        self._filenames.append(filename)
        return filename

    @property
    def filenames(self):
        """ Get a list of all generated filenames.
        """
        return self._filenames


class BandExpressionError(ValueError):
    pass


ALLOWED_NODE_TYPES = (
    _ast.Module,
    _ast.Expr,
    _ast.Load,
    _ast.Name,
    _ast.Call,

    _ast.UnaryOp,
    _ast.BinOp,

    _ast.Mult,
    _ast.Div,
    _ast.Add,
    _ast.Sub,
    _ast.Num if hasattr(_ast, 'Num') else _ast.Constant,

    _ast.BitAnd,
    _ast.BitOr,
    _ast.BitXor,

    _ast.USub,
)


def parse_expression(band_expression):
    """ Parse and validate the passed band expression
    """
    parsed = ast.parse(band_expression)
    for node in ast.walk(parsed):
        if not isinstance(node, ALLOWED_NODE_TYPES):
            raise BandExpressionError(
                'Invalid expression: %s' % type(node).__name__
            )
    return parsed.body[0].value


def parent_walk(node, depth=0):
    if depth == 0:
        yield None, node

    for child in ast.iter_child_nodes(node):
        yield node, child
        for child, ancestor in parent_walk(child, depth+1):
            yield child, ancestor


def extract_fields(band_expression):
    """ Extract the fields required to generate the output band.
        :param band_expression: the band expression to extract the fields of
        :type band_expression: str
        :return: a list of field names
        :rtype: list
    """
    if isinstance(band_expression, string_types):
        root_expr = parse_expression(band_expression)
    else:
        root_expr = band_expression
    return [
        node.id
        for parent, node in parent_walk(root_expr)
        if isinstance(node, _ast.Name) and not (
            isinstance(parent, _ast.Call) and parent.func == node
        )
    ]


class BrowseCreationInfo(object):
    def __init__(self, filename, env, bands=None):
        self.filename = filename
        self.env = env
        self.bands = bands


# make shortcut: if all relevant bands are in one dataset, we
# will only return that and give the band numbers
def single_file_and_indices(band_expressions, fields_and_coverages):
    band_indices = []
    filenames = set()
    env = None
    for band_expression in band_expressions:
        fields = extract_fields(band_expression)
        if len(fields) != 1:
            return None, None, None

        field = fields[0]

        coverages = fields_and_coverages[field]
        # break early if we are dealing with more than one coverage
        if len(coverages) != 1:
            return None, None, None

        coverage = coverages[0]
        location = coverage.get_location_for_field(field)

        filenames.add(location.path)
        band_indices.append(
            coverage.get_band_index_for_field(field)
        )
        env = location.env

    if len(filenames) == 1:
        return filenames.pop(), env, band_indices

    return None, None, None


def generate_browse(band_expressions, fields_and_coverages,
                    width, height, bbox, crs, generator=None):
    """ Produce a temporary VRT file describing how transformation of the
        coverages to browses.

        :param band_exressions: the band expressions for the various bands
        :param fields_and_coverages: a dictionary mapping the field names to
                                     all coverages with that field
        :param: band_expressiosn: list of strings
        :type fields_and_coverages: dict
        :return: A tuple of the filename of the output file and the generator
                 which was used to generate the filenames.
                 In most cases this is the filename refers to a generated VRT
                 file in very simple cases the file might actually refer to an
                 original file.
        :rtype: tuple
    """
    generator = generator or FilenameGenerator('/vsimem/{uuid}.vrt')

    # out_band_filenames = []

    parsed_expressions = [
        parse_expression(band_expression)
        for band_expression in band_expressions
    ]

    is_simple = all(isinstance(expr, _ast.Name) for expr in parsed_expressions)

    if not is_simple:
        return _generate_browse_complex(
            parsed_expressions, fields_and_coverages,
            width, height, bbox, crs, generator
        ), generator, True

    single_filename, env, bands = single_file_and_indices(
        band_expressions, fields_and_coverages
    )

    # for single files, we make a shortcut and just return it and the used
    # bands
    if single_filename:
        return (
            BrowseCreationInfo(single_filename, env, bands),
            generator, False
        )

    else:
        return _generate_browse_complex(
            parsed_expressions, fields_and_coverages,
            width, height, bbox, crs, generator
        ), generator, True

    # iterate over the input band expressions
    # for band_expression in band_expressions:
    #     fields = extract_fields(band_expression)

    #     selected_filenames = []

    #     # iterate over all fields that the output band shall be comprised of
    #     for field in fields:
    #         coverages = fields_and_coverages[field]

    #         # iterate over all coverages for that field to select the single
    #         # field
    #         for coverage in coverages:
    #             location = coverage.get_location_for_field(field)
    #             orig_filename = location.path
    #             orig_band_index = coverage.get_band_index_for_field(field)

    #             # only make a VRT to select the band if band count for the
    #             # dataset > 1
    #             if location.field_count == 1:
    #                 selected_filename = orig_filename
    #             else:
    #                 selected_filename = generator.generate()
    #                 vrt.select_bands(
    #                     orig_filename, location.env,
    #                     [orig_band_index], selected_filename
    #                 )

    #             selected_filenames.append(selected_filename)

    #     # if only a single file is required to generate the output band, return
    #     # it.
    #     if len(selected_filenames) == 1:
    #         out_band_filename = selected_filenames[0]

    #     # otherwise mosaic all the input bands to form a composite image
    #     else:
    #         out_band_filename = generator.generate()
    #         vrt.mosaic(selected_filenames, out_band_filename)

    #     out_band_filenames.append(out_band_filename)

    # # make shortcut here, when we only have one band, just return it
    # if len(out_band_filenames) == 1:
    #     return (
    #         BrowseCreationInfo(out_band_filenames[0], None), generator, False
    #     )

    # # return the stacked bands as a VRT
    # else:
    #     stacked_filename = generator.generate()
    #     vrt.stack_bands(out_band_filenames, env or location.env, stacked_filename)
    #     return (
    #         BrowseCreationInfo(stacked_filename, None), generator, False
    #     )


def _generate_browse_complex(parsed_exprs, fields_and_coverages,
                             width, height, bbox, crs, generator):
    o_x = bbox[0]
    o_y = bbox[3]
    res_x = (bbox[2] - bbox[0]) / width
    res_y = -(bbox[3] - bbox[1]) / height
    tiff_driver = gdal.GetDriverByName('GTiff')

    field_names = set()
    for parsed_expression in parsed_exprs:
        field_names |= set(extract_fields(parsed_expression))

    fields_and_datasets = {}
    for field_name in field_names:
        coverages = fields_and_coverages[field_name]
        field_data = warp_fields(
            coverages, field_name, bbox, crs, width, height
        )
        fields_and_datasets[field_name] = field_data

    out_filename = generator.generate('tif')
    tiff_driver = gdal.GetDriverByName('GTiff')
    out_ds = tiff_driver.Create(
        out_filename,
        width, height, len(parsed_exprs),
        gdal.GDT_Float32,
        options=[
            "TILED=YES",
            "COMPRESS=PACKBITS"
        ]
    )
    out_ds.SetGeoTransform([o_x, res_x, 0, o_y, 0, res_y])
    out_ds.SetProjection(osr.SpatialReference(crs).wkt)

    for band_index, parsed_expr in enumerate(parsed_exprs, start=1):
        with np.errstate(divide='ignore', invalid='ignore'):
            out_data = _evaluate_expression(
                parsed_expr, fields_and_datasets, generator
            )

        if isinstance(out_data, (int, float)):
            out_data = np.full((height, width), out_data)

        out_band = out_ds.GetRasterBand(band_index)
        out_band.WriteArray(out_data)

    return BrowseCreationInfo(out_filename, None)


operator_map = {
    _ast.Add: operator.add,
    _ast.Sub: operator.sub,
    _ast.Div: operator.truediv,
    _ast.Mult: operator.mul,
}

function_map = {
    'sin': np.sin,
    'cos': np.cos,
    'tan': np.tan,
    'arcsin': np.arcsin,
    'arccos': np.arccos,
    'arctan': np.arctan,
    'hypot': np.hypot,
    'arctan2': np.arctan2,
    'degrees': np.degrees,
    'radians': np.radians,
    'unwrap': np.unwrap,
    'deg2rad': np.deg2rad,
    'rad2deg': np.rad2deg,
    'sinh': np.sinh,
    'cosh': np.cosh,
    'tanh': np.tanh,
    'arcsinh': np.arcsinh,
    'arccosh': np.arccosh,
    'arctanh': np.arctanh,
    'exp': np.exp,
    'expm1': np.expm1,
    'exp2': np.exp2,
    'log': np.log,
    'log10': np.log10,
    'log2': np.log2,
    'log1p': np.log1p,
}


def _evaluate_expression(expr, fields_and_datasets, generator):
    if isinstance(expr, _ast.Name):
        return fields_and_datasets[expr.id]

    elif isinstance(expr, _ast.BinOp):
        left_data = _evaluate_expression(
            expr.left, fields_and_datasets, generator
        )

        right_data = _evaluate_expression(
            expr.right, fields_and_datasets, generator
        )

        op = operator_map[type(expr.op)]
        return op(left_data, right_data)

    elif isinstance(expr, _ast.Call):
        if not isinstance(expr.func, _ast.Name):
            raise BrowseGenerationError('Invalid function call')

        func = function_map.get(expr.func.id)
        if not func:
            raise BrowseGenerationError(
                'Invalid function %s, available functions are '
                % (expr.func.id, ', '.join(function_map.keys()))
            )

        if not len(expr.args) == 1:
            raise BrowseGenerationError(
                'Invalid number of arguments for function call'
            )

        args_data = [
            _evaluate_expression(
                arg, fields_and_datasets, generator
            ) for arg in expr.args
        ]
        res = func(*args_data)
        logger.info("%s, %s, %s" % (expr.func.id, np.min(res), np.max(res)))
        return res

    elif hasattr(_ast, 'Num') and isinstance(expr, _ast.Num):
        return expr.n

    elif hasattr(_ast, 'Constant') and isinstance(expr, _ast.Constant):
        return expr.value
